#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

##############################################
# These are part of the core.py Qiling class #
# handling hooks                             #
##############################################


from unicorn import *
from .utils import catch_KeyboardInterrupt
from .const import QL_HOOK_BLOCK
from .exception import QlErrorCoreHook

class Hook:
    def __init__(self, callback, user_data=None, begin=1, end=0):
        self.callback = callback
        self.user_data = user_data
        self.begin = begin
        self.end = end

    def bound_check(self, pc, size=1):
        return (self.end < self.begin) or (self.begin <= pc <= self.end) or (self.begin <= (pc + size - 1) <= self.end)


    def check(self, *args):
        return True
    

    def call(self, ql, *args):
        if self.user_data == None:
            return self.callback(ql, *args)
        return self.callback(ql, *args, self.user_data)


class HookAddr(Hook):
    def __init__(self, callback, address, user_data=None):
        super(HookAddr, self).__init__(callback, user_data, address, address)
        self.addr = address
    

    def call(self, ql, *args):
        if self.user_data == None:
            return self.callback(ql)
        return self.callback(ql, self.user_data)


class HookIntr(Hook):
    def __init__(self, callback, intno, user_data=None):
        super(HookIntr, self).__init__(callback, user_data, 0, -1)
        self.intno = intno
    

    def check(self, ql, intno):
        ql.log.debug("Received Interupt: %i Hooked Interupt: %i" % (intno, self.intno))
        if intno < 0 or self.intno == intno:
            return True
        return False


class HookRet:
    def __init__(self, ql, t, h):
        self._ql = ql
        self._t = t
        self._h = h
    

    def remove(self):
        self._ql.hook_del(self._t, self._h)

# Don't assume self is Qiling.
class QlCoreHooks:
    def __init__(self, uc):
        self._h_uc = uc
        self._hook = {}
        self._hook_fuc = {}
        self._insn_hook = {}
        self._insn_hook_fuc = {}
        self._addr_hook = {}
        self._addr_hook_fuc = {}
        
        self.hook_intr_fuc = None
        self.hook_insn_fuc = None
        self.hook_code_fuc = None
        self.hook_block_fuc = None
        self.hook_mem_read_unmapped_fuc = None
        self.hook_mem_write_unmapped_fuc = None
        self.hook_mem_fetch_unmapped_fuc = None
        self.hook_mem_read_prot_fuc = None
        self.hook_mem_write_prot_fuc = None
        self.hook_mem_fetch_prot_fuc = None
        self.hook_mem_read_fuc = None
        self.hook_mem_write_fuc = None
        self.hook_mem_fetch_fuc = None
        self.hook_mem_read_after_fuc = None
        self.hook_insn_invalid_fuc = None


    ########################
    # Callback definitions #
    ########################
    def _callback_type3(self, uc, intno, pack_data):
        ql, user_data, callback = pack_data     # unpack what we packed for hook_add()
        if user_data:
            return callback(ql, intno, user_data)
        return callback(ql, intno)              # callback does not require user_data
    

    def _hook_intr_cb(self, uc, intno, pack_data):
        ql, hook_type = pack_data
        catched = False
        if hook_type in self._hook.keys():
            for h in self._hook[hook_type]:
                if h.check(ql, intno):
                    catched = True
                    ret = h.call(ql, intno)
                    if isinstance(ret, int) == True and ret & QL_HOOK_BLOCK  != 0:
                        break
        
        if catched == False:
            raise QlErrorCoreHook("_hook_intr_cb : catched == False")
    

    def _hook_insn_cb(self, uc, *args):
        ql, hook_type = args[-1]

        if hook_type in self._insn_hook.keys():
            for h in self._insn_hook[hook_type]:
                if h.bound_check(ql.reg.arch_pc):
                    ret = h.call(ql, *args[ : -1])
                    if isinstance(ret, int) == True and ret & QL_HOOK_BLOCK  != 0:
                        break


    def _callback_type4(self, uc, addr, size, pack_data):
        ql, user_data, callback = pack_data
        if user_data:
            return callback(ql, addr, size, user_data)
        return callback(ql, addr, size)


    def _callback_type4a(self, uc, _addr, _size, pack_data):
        ql, user_data, callback = pack_data
        if user_data:
            return callback(ql, user_data)
        return callback(ql)


    def _hook_trace_cb(self, uc, addr, size, pack_data):
        ql, hook_type = pack_data
        if hook_type in self._hook.keys():
            for h in self._hook[hook_type]:
                if h.bound_check(ql.reg.arch_pc):
                    ret = h.call(ql, addr, size)
                    if isinstance(ret, int) == True and ret & QL_HOOK_BLOCK  != 0:
                        break


    def _callback_type6(self, uc, access, addr, size, value, pack_data):
        ql, user_data, callback = pack_data
        if user_data:
            return callback(ql, addr, size, value, user_data)
        return callback(ql, addr, size, value)


    def _hook_mem_cb(self, uc, access, addr, size, value, pack_data):
        ql, hook_type = pack_data
        handled = False
        if hook_type in self._hook.keys():
            for h in self._hook[hook_type]:
                if h.bound_check(addr, size):
                    handled = True
                    ret = h.call(ql, access, addr, size, value)
                    if isinstance(ret, int) == True and ret & QL_HOOK_BLOCK  != 0:
                        break
        
        if hook_type in (UC_HOOK_MEM_READ_UNMAPPED, UC_HOOK_MEM_WRITE_UNMAPPED, UC_HOOK_MEM_FETCH_UNMAPPED, UC_HOOK_MEM_READ_PROT, UC_HOOK_MEM_WRITE_PROT, UC_HOOK_MEM_FETCH_PROT):
            if handled == False:
                raise QlErrorCoreHook("_hook_mem_cb : handled == False")
        return True


    def _callback_x86_syscall(self, uc, pack_data):
        ql, user_data, callback = pack_data
        if user_data:
            return callback(ql, user_data)
        return callback(ql)


    def _hook_insn_invalid_cb(self, uc, pack_data):
        ql, hook_type = pack_data
        catched = False
        if hook_type in self._hook.keys():
            for h in self._hook[hook_type]:
                catched = True
                ret = h.call(ql)
                if isinstance(ret, int) == True and ret & QL_HOOK_BLOCK  != 0:
                    break
        
        if catched == False:
            raise QlErrorCoreHook("_hook_intr_invalid_cb : catched == False")


    def _hook_addr_cb(self, uc, addr, size, pack_data):
        ql, addr = pack_data
        if addr in self._addr_hook.keys():
            for h in self._addr_hook[addr]:
                ret = h.call(ql, addr, size)
                if isinstance(ret, int) == True and ret & QL_HOOK_BLOCK  != 0:
                    break

    ###############
    # Class Hooks #
    ###############
    def _ql_hook_internal(self, hook_type, callback, user_data=None, *args):
        _callback = (catch_KeyboardInterrupt(self))(callback)
        # pack user_data & callback for wrapper _callback
        return self._h_uc.hook_add(hook_type, _callback, (self, user_data), 1, 0, *args)


    def _ql_hook_addr_internal(self, callback, user_data, address):
        _callback = (catch_KeyboardInterrupt(self))(callback)
        # pack user_data & callback for wrapper _callback
        return self._h_uc.hook_add(UC_HOOK_CODE, _callback, (self, user_data), address, address)


    def _ql_hook(self, hook_type, h, *args):
        base_type = [
            UC_HOOK_INTR,
            UC_HOOK_INSN,
            UC_HOOK_CODE,
            UC_HOOK_BLOCK,
            UC_HOOK_MEM_READ_UNMAPPED,
            UC_HOOK_MEM_WRITE_UNMAPPED,
            UC_HOOK_MEM_FETCH_UNMAPPED,
            UC_HOOK_MEM_READ_PROT,
            UC_HOOK_MEM_WRITE_PROT,
            UC_HOOK_MEM_FETCH_PROT,
            UC_HOOK_MEM_READ,
            UC_HOOK_MEM_WRITE,
            UC_HOOK_MEM_FETCH,
            UC_HOOK_MEM_READ_AFTER,
            UC_HOOK_INSN_INVALID
            ]
        for t in base_type:
            if (t & hook_type) != 0:
                if t in (UC_HOOK_INTR, ):
                    if t not in self._hook_fuc.keys():
                        self._hook_fuc[t] = self._ql_hook_internal(t, self._hook_intr_cb, t)
                    
                    if t not in self._hook.keys():
                        self._hook[t] = []
                    self._hook[t].append(h)
                
                if t in (UC_HOOK_INSN, ):
                    ins_t = args[0]
                    if ins_t not in self._insn_hook_fuc.keys():
                        self._insn_hook_fuc[ins_t] = self._ql_hook_internal(t, self._hook_insn_cb, ins_t, *args)
                    
                    if ins_t not in self._insn_hook.keys():
                        self._insn_hook[ins_t] = []
                    self._insn_hook[ins_t].append(h)
                
                if t in (UC_HOOK_CODE, UC_HOOK_BLOCK):
                    if t not in self._hook_fuc.keys():
                        self._hook_fuc[t] = self._ql_hook_internal(t, self._hook_trace_cb, t)

                    if t not in self._hook.keys():
                        self._hook[t] = []
                    self._hook[t].append(h)
                
                if t in (UC_HOOK_MEM_READ_UNMAPPED, UC_HOOK_MEM_WRITE_UNMAPPED, UC_HOOK_MEM_FETCH_UNMAPPED, UC_HOOK_MEM_READ_PROT, UC_HOOK_MEM_WRITE_PROT, UC_HOOK_MEM_FETCH_PROT, UC_HOOK_MEM_READ, UC_HOOK_MEM_WRITE, UC_HOOK_MEM_FETCH, UC_HOOK_MEM_READ_AFTER):
                    if t not in self._hook_fuc.keys():
                        self._hook_fuc[t] = self._ql_hook_internal(t, self._hook_mem_cb, t)

                    if t not in self._hook.keys():
                        self._hook[t] = []
                    self._hook[t].append(h)
                
                if t in (UC_HOOK_INSN_INVALID, ):
                    if t not in self._hook_fuc.keys():
                        self._hook_fuc[t] = self._ql_hook_internal(t, self._hook_insn_invalid_cb, t)

                    if t not in self._hook.keys():
                        self._hook[t] = []
                    self._hook[t].append(h)


    def ql_hook(self, hook_type, callback, user_data=None, begin=1, end=0, *args):
        h = Hook(callback, user_data, begin, end)
        self._ql_hook(hook_type, h, *args)
        return HookRet(self, hook_type, h)


    def hook_code(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_CODE, callback, user_data, begin, end)


    def hook_intr(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_INTR,  callback, user_data, begin, end)


    def hook_block(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_BLOCK, callback, user_data, begin, end)


    def hook_mem_unmapped(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_UNMAPPED, callback, user_data, begin, end)


    def hook_mem_read_invalid(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_READ_INVALID, callback, user_data, begin, end)


    def hook_mem_write_invalid(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_WRITE_INVALID, callback, user_data, begin, end)


    def hook_mem_fetch_invalid(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_FETCH_INVALID, callback, user_data, begin, end)


    def hook_mem_valid(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_VALID, callback, user_data, begin, end)
    
    def hook_mem_invalid(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_INVALID, callback, user_data, begin, end)


    # a convenient API to set callback for a single address
    def hook_address(self, callback, address, user_data=None):
        h = HookAddr(callback, address, user_data)

        if address not in self._addr_hook_fuc.keys():
            self._addr_hook_fuc[address] = self._ql_hook_addr_internal(self._hook_addr_cb, address, address)

        if address not in self._addr_hook.keys():
            self._addr_hook[address] = []

        self._addr_hook[address].append(h)
        return HookRet(self, None, h)
    

    def hook_intno(self, callback, intno, user_data=None):
        h = HookIntr(callback, intno, user_data)
        self._ql_hook(UC_HOOK_INTR, h)
        return HookRet(self, UC_HOOK_INTR, h)


    def hook_mem_read(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_READ, callback, user_data, begin, end)


    def hook_mem_write(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_WRITE, callback, user_data, begin, end)


    def hook_mem_fetch(self, callback, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_MEM_FETCH, callback, user_data, begin, end)


    def hook_insn(self, callback, arg1, user_data=None, begin=1, end=0):
        return self.ql_hook(UC_HOOK_INSN, callback, user_data, begin, end, arg1)


    def hook_del(self, *args):
        if len(args) != 1 and len(args) != 2:
            return

        if isinstance(args[0], HookRet):
            args[0].remove()
            return
        else:
            hook_type, h = args

        base_type = [
            UC_HOOK_INTR,
            UC_HOOK_INSN,
            UC_HOOK_CODE,
            UC_HOOK_BLOCK,
            UC_HOOK_MEM_READ_UNMAPPED,
            UC_HOOK_MEM_WRITE_UNMAPPED,
            UC_HOOK_MEM_FETCH_UNMAPPED,
            UC_HOOK_MEM_READ_PROT,
            UC_HOOK_MEM_WRITE_PROT,
            UC_HOOK_MEM_FETCH_PROT,
            UC_HOOK_MEM_READ,
            UC_HOOK_MEM_WRITE,
            UC_HOOK_MEM_FETCH,
            UC_HOOK_MEM_READ_AFTER,
            UC_HOOK_INSN_INVALID
            ]
        if isinstance(h, HookAddr):
            if h.addr in self._addr_hook.keys():
                if h in self._addr_hook[h.addr]:
                    del self._addr_hook[h.addr][self._addr_hook[h.addr].index(h)]

                    if len(self._addr_hook[h.addr]) == 0:
                        self._h_uc.hook_del(self._addr_hook_fuc[h.addr])
                        del self._addr_hook_fuc[h.addr]
            
            return

        for t in base_type:
            if (t & hook_type) != 0:
                if t in (UC_HOOK_INTR, UC_HOOK_CODE, UC_HOOK_BLOCK, UC_HOOK_MEM_READ_UNMAPPED, UC_HOOK_MEM_WRITE_UNMAPPED, UC_HOOK_MEM_FETCH_UNMAPPED, UC_HOOK_MEM_READ_PROT, UC_HOOK_MEM_WRITE_PROT, UC_HOOK_MEM_FETCH_PROT, UC_HOOK_MEM_READ, UC_HOOK_MEM_WRITE, UC_HOOK_MEM_FETCH, UC_HOOK_MEM_READ_AFTER, UC_HOOK_INSN_INVALID):
                    if t in self._hook.keys():
                        if h in self._hook[t]:
                            del self._hook[t][self._hook[t].index(h)]

                            if len(self._hook[t]) == 0:
                                self._h_uc.hook_del(self._hook_fuc[t])
                                del self._hook_fuc[t]
                
                if t in (UC_HOOK_INSN, ):
                    if t in self._insn_hook.keys():
                        if h in self._insn_hook[t]:
                            del self._insn_hook[t][self._insn_hook[t].index(h)]

                            if len(self._insn_hook[t]) == 0:
                                self._h_uc.hook_del(self._insn_hook_fuc[t])
                                del self._insn_hook_fuc[t]
                    

    def clear_hooks(self):
        for i in self._hook_fuc.keys():
            self._h_uc.hook_del(self._hook_fuc[i])
        
        for i in self._insn_hook_fuc.keys():
            self._h_uc.hook_del(self._insn_hook_fuc[i])

        for i in self._addr_hook_fuc.keys():
            self._h_uc.hook_del(self._addr_hook_fuc[i])

        self.clear_ql_hooks()
    
    
    def clear_ql_hooks(self):
        self._hook = {}
        self._hook_fuc = {}
        self._insn_hook = {}
        self._insn_hook_fuc = {}
        self._addr_hook = {}
        self._addr_hook_fuc = {}



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

"""
This module is intended for general purpose functions that are only used in qiling.os
"""

import ctypes, inspect, os, struct, uuid

from json import dumps
from pathlib import Path, PurePosixPath, PureWindowsPath, PosixPath, WindowsPath
from unicorn import *
from unicorn.arm_const import *
from unicorn.x86_const import *
from unicorn.arm64_const import *
from unicorn.mips_const import *
from capstone import *
from capstone.arm_const import *
from capstone.x86_const import *
from capstone.arm64_const import *
from capstone.mips_const import *
from keystone import *


from qiling.const import *
from qiling.exception import *
from .const import *
from qiling.os.windows.wdk_const import *
from qiling.os.windows.structs import *
from qiling.utils import verify_ret

# OH-MY-WIN32 !!!
# Some codes from cygwin.
class PathUtils:

    # Basic guide:
    #     We should only handle "normal" paths like "C:\Windows\System32" and "bin/a.exe" for users.
    #     For UNC paths like '\\.\PHYSICALDRIVE0" and "\\Server\Share", they should be implemented 
    #     by users via fs mapping interface.
    @staticmethod
    def convert_win32_to_posix(rootfs, cwd, path):
        # rootfs is a concrete path.
        rootfs = Path(rootfs)
        # cwd and path are pure paths
        cwd = PurePosixPath(cwd[1:])

        result = None
        # Things are complicated here.
        # See https://docs.microsoft.com/zh-cn/windows/win32/fileio/naming-a-file?redirectedfrom=MSDN
        if PureWindowsPath(path).is_absolute():
            if (len(path) >= 2 and path[0] == '\\' and path[1] == '\\') or \
                (len(path) >= 3 and path[0].isalpha() and path[2] == '\\'): # \\.\PhysicalDrive0 or \\Server\Share\Directory or X:\
                # UNC path should be handled in fs mapping. If not, append it to rootfs directly.
                pw = PureWindowsPath(path)
                result = rootfs / PathUtils.normalize(pw)
            else:
                # code should never reach here.
                result = rootfs / PathUtils.normalize(path)
        else:
            if len(path) >= 3 and path[:3] == r'\\?' or path[:3] == r'\??': # \??\ or \\?\ or \Device\..
                # Similair to \\.\, it should be handled in fs mapping.
                pw = PureWindowsPath(path)
                result = rootfs / PathUtils.normalize(cwd / pw.relative_to(pw.anchor).as_posix())
            else:
                # a normal relative path
                result = rootfs / PathUtils.normalize(cwd / PureWindowsPath(path).as_posix())
        return result


    @staticmethod
    def convert_posix_to_win32(rootfs, cwd, path):
        # rootfs is a concrete path.
        rootfs = Path(rootfs)
        # cwd and path are pure paths
        cwd = PurePosixPath(cwd[1:])
        path = PurePosixPath(path)
        if path.is_absolute():
            return rootfs / PathUtils.normalize(path)
        else:
            return rootfs / PathUtils.normalize(cwd / path)
    
    @staticmethod
    def convert_for_native_os(rootfs, cwd, path):
        rootfs = Path(rootfs)
        cwd = PurePosixPath(cwd[1:])
        path = Path(path)
        if path.is_absolute():
            return rootfs / PathUtils.normalize(path)
        else:
            return rootfs / PathUtils.normalize(cwd / path.as_posix())

    @staticmethod
    def normalize(path):
        if type(path) is PurePosixPath:
            normalized_path = PurePosixPath()
        elif type(path) is PureWindowsPath:
            normalized_path = PureWindowsPath()
        elif type(path) is PosixPath:
            normalized_path = PosixPath()
        elif type(path) is WindowsPath:
            normalized_path = WindowsPath()

        # remove anchor (necessary for Windows UNC paths) and convert to relative path
        if path.is_absolute():
            path = path.relative_to(path.anchor)

        for p in path.parts:
            if p == '.':
                continue

            if p == '..':
                normalized_path = normalized_path.parent
                continue

            normalized_path /= p

        return normalized_path

class QlOsUtils:
    def __init__(self, ql):
        self.ql = ql
        self.path = None
        self.md = None
        self._disasm_hook = None
        self._block_hook = None


    def string_appearance(self, string):
        strings = string.split(" ")
        for string in strings:
            val = self.appeared_strings.get(string, set())
            val.add(self.syscalls_counter)
            self.appeared_strings[string] = val


    def read_wstring(self, address):
        result = ""
        char = self.ql.mem.read(address, 2)
        while char.decode(errors="ignore") != "\x00\x00":
            address += 2
            result += char.decode(errors="ignore")
            char = self.ql.mem.read(address, 2)
        # We need to remove \x00 inside the string. Compares do not work otherwise
        result = result.replace("\x00", "")
        self.string_appearance(result)
        return result


    def read_cstring(self, address):
        result = ""
        char = self.ql.mem.read(address, 1)
        while char.decode(errors="ignore") != "\x00":
            address += 1
            result += char.decode(errors="ignore")
            char = self.ql.mem.read(address, 1)
        self.string_appearance(result)
        return result

    def print_function(self, address, function_name, params, ret, passthru=False):
        PRINTK_LEVEL = {
            0: 'KERN_EMERGE',
            1: 'KERN_ALERT',
            1: 'KERN_CRIT',
            2: 'KERN_INFO',
            3: 'KERN_ERR',
            4: 'KERN_WARNING',
            5: 'KERN_NOTICE',
            6: 'KERN_INFO',
            7: 'KERN_DEBUG',
            8: '',
            9: 'KERN_CONT',
        }
        
        if function_name.startswith('hook_'):
            function_name = function_name[5:]

        if function_name in ("__stdio_common_vfprintf", "__stdio_common_vfwprintf", "printf", "wsprintfW", "sprintf"):
            return
        
        def _parse_param(param):
            name, value = param

            if type(value) is str:
                return f'{name:s} = "{value}"'
            elif type(value) is bytearray:
                return f'{name:s} = "{value.decode("utf-8")}"'
            elif type(value) is tuple:
                # we just need the string, not the address in the log
                return f'{name:s} = "{value[1]}"'

            # default to hexadecimal representation
            return f'{name:s} = {value:#x}'

        # arguments list
        fargs = (_parse_param(param) for param in params.items())

        # optional suffixes: return value and passthrough
        fret = f' = {ret:#x}' if ret is not None else ''
        fpass = f' (PASSTHRU)' if passthru else ''

        #TODO: Old code from demigod, ready to cleanup
        if self.ql.ostype in QL_OS_POSIX and self.ql.loader.is_driver:
            log = '0x%0.2x: %s(' % (address, function_name)
            for each in params:
                value = params[each]
                if type(value) == str or type(value) == bytearray:
                    if function_name == 'printk':
                        info = value[:2]
                        try:
                            level = PRINTK_LEVEL[int(info[1])]
                            value = value[2:]
                            log += '%s = %s "%s", ' %(each, level, value)
                        except:
                            log += '%s = "%s", ' %(each, value)
                    else:
                        log += '%s = "%s", ' %(each, value)
                elif type(value) == tuple:
                    log += '%s = 0x%x, ' % (each, value[0])
                else:
                    log += '%s = 0x%x, ' % (each, value)
            log = log.strip(", ")
            log += ')'
            if ret is not None:
                # do not print result for printk()
                if function_name != 'printk':
                    log += ' = 0x%x' % ret
        else:    
            log = f'0x{address:02x}: {function_name:s}({", ".join(fargs)}){fret}{fpass}'

        if self.ql.output == QL_OUTPUT.DEBUG:
            self.ql.log.debug(log)
        else:
            log = log.partition(" ")[-1]
            self.ql.log.info(log)

    def vprintf(self, address, fmt, params_addr, name, wstring=False):
        count = fmt.count("%")
        params = []
        if count > 0:
            for i in range(count):
                param = self.ql.mem.read(params_addr + i * self.ql.pointersize, self.ql.pointersize)
                params.append(
                    self.ql.unpack(param)
                )
        return self.printf(address, fmt, params, name, wstring)

    def printf(self, address, fmt, params, name, wstring=False):
        if len(params) > 0:
            formats = fmt.split("%")[1:]
            index = 0
            for f in formats:
                if f.startswith("s"):
                    if wstring:
                        params[index] = self.read_wstring(params[index])
                    else:
                        params[index] = self.read_cstring(params[index])
                index += 1

            output = '%s(format = %s' % (name, repr(fmt))
            for each in params:
                if type(each) == str:
                    output += ', "%s"' % each
                else:
                    output += ', 0x%0.2x' % each
            output += ')'
            fmt = fmt.replace("%llx", "%x")
            stdout = fmt % tuple(params)
            output += " = 0x%x" % len(stdout)
        else:
            output = '%s(format = %s) = 0x%x' % (name, repr(fmt), len(fmt))
            stdout = fmt
        self.ql.log.info(output)
        self.ql.os.stdout.write(bytes(stdout, 'utf-8'))
        return len(stdout), stdout

    def lsbmsb_convert(self, sc, size=4):
        split_bytes = []
        n = size
        for index in range(0, len(sc), n):
            split_bytes.append((sc[index: index + n])[::-1])

        ebsc = b""
        for i in split_bytes:
            ebsc += i

        return ebsc

    def convert_path(self, rootfs, cwd, path):
        if  (self.ql.ostype == self.ql.platform ) \
            or (self.ql.ostype in [QL_OS.LINUX, QL_OS.MACOS] and self.ql.platform in [QL_OS.LINUX, QL_OS.MACOS]):
            return PathUtils.convert_for_native_os(rootfs, cwd, path)
        elif self.ql.ostype in [QL_OS.LINUX, QL_OS.MACOS] and self.ql.platform == QL_OS.WINDOWS:
            return PathUtils.convert_posix_to_win32(rootfs, cwd, path)
        elif self.ql.ostype == QL_OS.WINDOWS and self.ql.platform in [QL_OS.LINUX, QL_OS.MACOS]:
            return PathUtils.convert_win32_to_posix(rootfs, cwd, path)
        else:
            return None
    
    def transform_to_link_path(self, path):
        if self.ql.multithread:
            cur_path = self.ql.os.thread_management.cur_thread.current_path
        else:
            cur_path = self.ql.os.current_path

        # Sanity check.
        if cur_path[0] != '/':
            self.ql.log.info(f"Warning: cur_path doesn't start with a /")
        
        rootfs = self.ql.rootfs
        real_path  = self.convert_path(rootfs, cur_path, path)

        return str(real_path.absolute())

    def transform_to_real_path(self, path):
        from types import FunctionType

        if self.ql.multithread:
            cur_path = self.ql.os.thread_management.cur_thread.current_path
        else:
            cur_path = self.ql.os.current_path

        # Sanity check.
        if cur_path[0] != '/':
            self.ql.log.info(f"Warning: cur_path must start with /")

        rootfs = self.ql.rootfs
        real_path = self.convert_path(rootfs, cur_path, path)
        
        if os.path.islink(real_path):
            link_path = Path(os.readlink(real_path))
            if not link_path.is_absolute():
                real_path = Path(os.path.join(os.path.dirname(real_path), link_path))
            
        return str(real_path.absolute())

    # The `relative path` here refers to the path which is relative to the rootfs.
    def transform_to_relative_path(self, path):
        if self.ql.multithread:
            cur_path = self.ql.os.thread_management.cur_thread.current_path
        else:
            cur_path = self.ql.os.current_path

        return str(Path(cur_path) / path)

    def post_report(self):
        self.ql.log.debug("Syscalls called")
        for key, values in self.ql.os.syscalls.items():
            self.ql.log.debug("%s:" % key)
            for value in values:
                self.ql.log.debug("%s " % str(dumps(value)))
        self.ql.log.debug("Registries accessed")
        for key, values in self.ql.os.registry_manager.accessed.items():
            self.ql.log.debug("%s:" % key)
            for value in values:
                self.ql.log.debug("%s " % str(dumps(value)))
        self.ql.log.debug("Strings")
        for key, values in self.ql.os.appeared_strings.items():
            val = " ".join([str(word) for word in values])
            self.ql.log.debug("%s: %s" % (key, val))


    def exec_arbitrary(self, start, end):
        old_sp = self.ql.reg.arch_sp

        # we read where this hook is supposed to return
        ret = self.ql.stack_read(0)

        def restore(ql):
            self.ql.log.debug(f"Executed code from 0x{start:x} to 0x{end:x}")
            # now we can restore the register to be where we were supposed to
            old_hook_addr = ql.reg.arch_pc
            ql.reg.arch_sp = old_sp + (ql.archbit // 8)
            ql.reg.arch_pc = ret
            # we want to execute the code once, not more
            ql.hook_address(lambda q: None, old_hook_addr)

        # we have to set an address to restore the registers
        self.ql.hook_address(restore, end, )
        # we want to rewrite the return address to the function
        self.ql.stack_write(0, start)

    def get_offset_and_name(self, addr):
        for begin, end, access, name in self.ql.mem.map_info:
            if begin <= addr and end > addr:
                return addr-begin, name
        return addr, '-'

    def disassembler(self, ql, address, size):
        tmp = self.ql.mem.read(address, size)

        if not self.md:
            self.md = self.ql.create_disassembler()
        elif self.ql.archtype == QL_ARCH.ARM: # Update disassembler for arm considering thumb swtich.
            self.md = self.ql.create_disassembler()

        insn = self.md.disasm(tmp, address)
        opsize = int(size)

        offset, name = self.get_offset_and_name(address)
        log_data = '0x%0*x {%-20s + 0x%06x}   ' % (self.ql.archbit // 4, address, name, offset)

        temp_str = ""
        for i in tmp:
            temp_str += ("%02x " % i)
        log_data += temp_str.ljust(30)

        first = True
        for i in insn:
            if not first:
                log_data += '\n> '
            first = False
            log_data += "%s %s" % (i.mnemonic, i.op_str)
        self.ql.log.info(log_data)

        if self.ql.output == QL_OUTPUT.DUMP:
            for reg in self.ql.reg.register_mapping:
                if isinstance(reg, str):
                    REG_NAME = reg
                    REG_VAL = self.ql.reg.read(reg)
                    self.ql.log.debug("%s\t:\t 0x%x" % (REG_NAME, REG_VAL))

    def setup_output(self):
        def ql_hook_block_disasm(ql, address, size):
            self.ql.log.info("\nTracing basic block at 0x%x" % (address))

        if self._disasm_hook:
            self._disasm_hook.remove()
            self._disasm_hook = None
        if self._block_hook:
            self._block_hook.remove()
            self._block_hook = None

        if self.ql.output in (QL_OUTPUT.DISASM, QL_OUTPUT.DUMP):
            if self.ql.output == QL_OUTPUT.DUMP:
                self._block_hook = self.ql.hook_block(ql_hook_block_disasm)
            self._disasm_hook = self.ql.hook_code(self.disassembler)

    def stop(self):
        if self.ql.multithread:
            td = self.thread_management.stop() 
        else:
            self.ql.emu_stop()

    def read_guid(self, address):
        result = ""
        raw_guid = self.ql.mem.read(address, 16)
        return uuid.UUID(bytes_le=bytes(raw_guid))

    def io_Write(self, in_buffer):
        if self.ql.ostype == QL_OS.WINDOWS:

            if self.ql.loader.driver_object.MajorFunction[IRP_MJ_WRITE] == 0:
                # raise error?
                return (False, None)

        if self.ql.archbit == 32:
            buf = self.ql.mem.read(self.ql.loader.driver_object.DeviceObject, ctypes.sizeof(DEVICE_OBJECT32))
            device_object = DEVICE_OBJECT32.from_buffer(buf)
        else:
            buf = self.ql.mem.read(self.ql.loader.driver_object.DeviceObject, ctypes.sizeof(DEVICE_OBJECT64))
            device_object = DEVICE_OBJECT64.from_buffer(buf)

        alloc_addr = []
        def build_mdl(buffer_size, data=None):
            if self.ql.archtype == QL_ARCH.X8664:
                mdl = MDL64()
            else:
                mdl = MDL32()

            mapped_address = self.heap.alloc(buffer_size)
            alloc_addr.append(mapped_address)
            mdl.MappedSystemVa.value = mapped_address
            mdl.StartVa.value = mapped_address
            mdl.ByteOffset = 0
            mdl.ByteCount = buffer_size
            if data:
                written = data if len(data) <= buffer_size else data[:buffer_size]
                self.ql.mem.write(mapped_address, written)

            return mdl
        # allocate memory regions for IRP and IO_STACK_LOCATION
        if self.ql.archtype == QL_ARCH.X8664:
            irp_addr = self.heap.alloc(ctypes.sizeof(IRP64))
            alloc_addr.append(irp_addr)
            irpstack_addr = self.heap.alloc(ctypes.sizeof(IO_STACK_LOCATION64))
            alloc_addr.append(irpstack_addr)
            # setup irp stack parameters
            irpstack = IO_STACK_LOCATION64()
            # setup IRP structure
            irp = IRP64()
            irp.irpstack = ctypes.cast(irpstack_addr, ctypes.POINTER(IO_STACK_LOCATION64))
        else:
            irp_addr = self.heap.alloc(ctypes.sizeof(IRP32))
            alloc_addr.append(irp_addr)
            irpstack_addr = self.heap.alloc(ctypes.sizeof(IO_STACK_LOCATION32))
            alloc_addr.append(irpstack_addr)
            # setup irp stack parameters
            irpstack = IO_STACK_LOCATION32()
            # setup IRP structure
            irp = IRP32()
            irp.irpstack = ctypes.cast(irpstack_addr, ctypes.POINTER(IO_STACK_LOCATION32))

        irpstack.MajorFunction = IRP_MJ_WRITE
        irpstack.Parameters.Write.Length = len(in_buffer)
        self.ql.mem.write(irpstack_addr, bytes(irpstack))

        if device_object.Flags & DO_BUFFERED_IO:
            # BUFFERED_IO
            system_buffer_addr = self.heap.alloc(len(in_buffer))
            alloc_addr.append(system_buffer_addr)
            self.ql.mem.write(system_buffer_addr, bytes(in_buffer))
            irp.AssociatedIrp.SystemBuffer.value = system_buffer_addr
        elif device_object.Flags & DO_DIRECT_IO:
            # DIRECT_IO
            mdl = build_mdl(len(in_buffer))
            if self.ql.archtype == QL_ARCH.X8664:
                mdl_addr = self.heap.alloc(ctypes.sizeof(MDL64))
            else:
                mdl_addr = self.heap.alloc(ctypes.sizeof(MDL32))

            alloc_addr.append(mdl_addr)

            self.ql.mem.write(mdl_addr, bytes(mdl))
            irp.MdlAddress.value = mdl_addr
        else:
            # NEITHER_IO
            input_buffer_size = len(in_buffer)
            input_buffer_addr = self.heap.alloc(input_buffer_size)
            alloc_addr.append(input_buffer_addr)
            self.ql.mem.write(input_buffer_addr, bytes(in_buffer))
            irp.UserBuffer.value = input_buffer_addr

        # everything is done! Write IRP to memory
        self.ql.mem.write(irp_addr, bytes(irp))

        # set function args
        self.set_function_args((self.ql.loader.driver_object.DeviceObject, irp_addr))

        try:
            # now emulate 
            self.ql.run(self.ql.loader.driver_object.MajorFunction[IRP_MJ_WRITE])
        except UcError as err:
            verify_ret(self.ql, err)
            
        # read current IRP state
        if self.ql.archtype == QL_ARCH.X8664:
            irp_buffer = self.ql.mem.read(irp_addr, ctypes.sizeof(IRP64))
            irp = IRP64.from_buffer(irp_buffer)
        else:
            irp_buffer = self.ql.mem.read(irp_addr, ctypes.sizeof(IRP32))
            irp = IRP32.from_buffer(irp_buffer)

        io_status = irp.IoStatus
        # now free all alloc memory
        for addr in alloc_addr:
            # print("freeing heap memory at 0x%x" %addr) # FIXME: the output is not deterministic??
            self.heap.free(addr)
        return True, io_status.Information.value

    # Emulate DeviceIoControl() of Windows
    # BOOL DeviceIoControl(
    #      HANDLE       hDevice,
    #      DWORD        dwIoControlCode,
    #      LPVOID       lpInBuffer,
    #      DWORD        nInBufferSize,
    #      LPVOID       lpOutBuffer,
    #      DWORD        nOutBufferSize,
    #      LPDWORD      lpBytesReturned,
    #      LPOVERLAPPED lpOverlapped);
    def ioctl(self, params):
        def ioctl_code(DeviceType, Function, Method, Access):
            return (DeviceType << 16) | (Access << 14) | (Function << 2) | Method

        alloc_addr = []
        def build_mdl(buffer_size, data=None):
            if self.ql.archtype == QL_ARCH.X8664:
                mdl = MDL64()
            else:
                mdl = MDL32()

            mapped_address = self.heap.alloc(buffer_size)
            alloc_addr.append(mapped_address)
            mdl.MappedSystemVa.value = mapped_address
            mdl.StartVa.value = mapped_address
            mdl.ByteOffset = 0
            mdl.ByteCount = buffer_size
            if data:
                written = data if len(data) <= buffer_size else data[:buffer_size]
                self.ql.mem.write(mapped_address, written)

            return mdl

        # quick simple way to manage all alloc memory
        if self.ql.ostype == QL_OS.WINDOWS:
            # print("DeviceControl callback is at 0x%x" %self.loader.driver_object.MajorFunction[IRP_MJ_DEVICE_CONTROL])
            if self.ql.loader.driver_object.MajorFunction[IRP_MJ_DEVICE_CONTROL] == 0:
                # raise error?
                return (None, None, None)

            # create new memory region to store input data
            _ioctl_code, output_buffer_size, in_buffer = params
            # extract data transfer method
            devicetype, function, ctl_method, access = _ioctl_code

            input_buffer_size = len(in_buffer)
            input_buffer_addr = self.heap.alloc(input_buffer_size)
            alloc_addr.append(input_buffer_addr)
            self.ql.mem.write(input_buffer_addr, bytes(in_buffer))

            # create new memory region to store out data
            output_buffer_addr = self.heap.alloc(output_buffer_size)
            alloc_addr.append(output_buffer_addr)

            # allocate memory regions for IRP and IO_STACK_LOCATION
            if self.ql.archtype == QL_ARCH.X8664:
                irp_addr = self.heap.alloc(ctypes.sizeof(IRP64))
                alloc_addr.append(irp_addr)
                irpstack_addr = self.heap.alloc(ctypes.sizeof(IO_STACK_LOCATION64))
                alloc_addr.append(irpstack_addr)
                # setup irp stack parameters
                irpstack = IO_STACK_LOCATION64()
                # setup IRP structure
                irp = IRP64()
                irp.irpstack = ctypes.cast(irpstack_addr, ctypes.POINTER(IO_STACK_LOCATION64))
            else:
                irp_addr = self.heap.alloc(ctypes.sizeof(IRP32))
                alloc_addr.append(irp_addr)
                irpstack_addr = self.heap.alloc(ctypes.sizeof(IO_STACK_LOCATION32))
                alloc_addr.append(irpstack_addr)
                # setup irp stack parameters
                irpstack = IO_STACK_LOCATION32()
                # setup IRP structure
                irp = IRP32()
                irp.irpstack = ctypes.cast(irpstack_addr, ctypes.POINTER(IO_STACK_LOCATION32))

                #print("32 stack location size = 0x%x" %ctypes.sizeof(IO_STACK_LOCATION32))
                #print("32 status block size = 0x%x" %ctypes.sizeof(IO_STATUS_BLOCK32))
                #print("32 irp size = 0x%x" %ctypes.sizeof(IRP32))
                #print("32 IoStatus offset = 0x%x" %IRP32.IoStatus.offset)
                #print("32 UserIosb offset = 0x%x" %IRP32.UserIosb.offset)
                #print("32 UserEvent offset = 0x%x" %IRP32.UserEvent.offset)
                #print("32 UserBuffer offset = 0x%x" %IRP32.UserBuffer.offset)
                #print("32 irpstack offset = 0x%x" %IRP32.irpstack.offset)
                #print("irp at %x, irpstack at %x" %(irp_addr, irpstack_addr))

            self.ql.log.info("IRP is at 0x%x, IO_STACK_LOCATION is at 0x%x" %(irp_addr, irpstack_addr))

            irpstack.Parameters.DeviceIoControl.IoControlCode = ioctl_code(devicetype, function, ctl_method, access)
            irpstack.Parameters.DeviceIoControl.OutputBufferLength = output_buffer_size
            irpstack.Parameters.DeviceIoControl.InputBufferLength = input_buffer_size
            irpstack.Parameters.DeviceIoControl.Type3InputBuffer.value = input_buffer_addr # used by IOCTL_METHOD_NEITHER
            self.ql.mem.write(irpstack_addr, bytes(irpstack))

            if ctl_method == METHOD_NEITHER:
                irp.UserBuffer.value = output_buffer_addr  # used by IOCTL_METHOD_NEITHER

            # allocate memory for AssociatedIrp.SystemBuffer
            # used by IOCTL_METHOD_IN_DIRECT, IOCTL_METHOD_OUT_DIRECT and IOCTL_METHOD_BUFFERED
            system_buffer_size = max(input_buffer_size, output_buffer_size)
            system_buffer_addr = self.heap.alloc(system_buffer_size)
            alloc_addr.append(system_buffer_addr)

            # init data from input buffer
            self.ql.mem.write(system_buffer_addr, bytes(in_buffer))
            irp.AssociatedIrp.SystemBuffer.value = system_buffer_addr

            if ctl_method in (METHOD_IN_DIRECT, METHOD_OUT_DIRECT):
                # Create MDL structure for output data
                # used by both IOCTL_METHOD_IN_DIRECT and IOCTL_METHOD_OUT_DIRECT
                mdl = build_mdl(output_buffer_size)
                if self.ql.archtype == QL_ARCH.X8664:
                    mdl_addr = self.heap.alloc(ctypes.sizeof(MDL64))
                else:
                    mdl_addr = self.heap.alloc(ctypes.sizeof(MDL32))

                alloc_addr.append(mdl_addr)

                self.ql.mem.write(mdl_addr, bytes(mdl))
                irp.MdlAddress.value = mdl_addr

            # everything is done! Write IRP to memory
            self.ql.mem.write(irp_addr, bytes(irp))

            # set function args
            self.ql.log.info("Executing IOCTL with DeviceObject = 0x%x, IRP = 0x%x" %(self.ql.loader.driver_object.DeviceObject, irp_addr))
            self.set_function_args((self.ql.loader.driver_object.DeviceObject, irp_addr))

            try:
                # now emulate IOCTL's DeviceControl
                self.ql.run(self.ql.loader.driver_object.MajorFunction[IRP_MJ_DEVICE_CONTROL])
            except UcError as err:
                verify_ret(self.ql, err)

            # read current IRP state
            if self.ql.archtype == QL_ARCH.X8664:
                irp_buffer = self.ql.mem.read(irp_addr, ctypes.sizeof(IRP64))
                irp = IRP64.from_buffer(irp_buffer)
            else:
                irp_buffer = self.ql.mem.read(irp_addr, ctypes.sizeof(IRP32))
                irp = IRP32.from_buffer(irp_buffer)

            io_status = irp.IoStatus

            # read output data
            output_data = b''
            if io_status.Status.Status >= 0:
                if ctl_method == METHOD_BUFFERED:
                    output_data = self.ql.mem.read(system_buffer_addr, io_status.Information.value)
                if ctl_method in (METHOD_IN_DIRECT, METHOD_OUT_DIRECT):
                    output_data = self.ql.mem.read(mdl.MappedSystemVa.value, io_status.Information.value)
                if ctl_method == METHOD_NEITHER:
                    output_data = self.ql.mem.read(output_buffer_addr, io_status.Information.value)

            # now free all alloc memory
            for addr in alloc_addr:
                # print("freeing heap memory at 0x%x" %addr) # FIXME: the output is not deterministic??
                self.heap.free(addr)
            #print("\n")

            return io_status.Status.Status, io_status.Information.value, output_data
        else: # TODO: IOCTL for non-Windows.
            pass        



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from enum import IntEnum

class QL_ENDIAN(IntEnum):
    EL = 1
    EB = 2


class QL_ARCH(IntEnum):
    X86 = 101
    X8664 = 102
    ARM = 103
    ARM_THUMB = 104
    ARM64 = 105
    MIPS = 106
    A8086 = 7


class QL_OS(IntEnum):
    LINUX = 201
    FREEBSD = 202
    MACOS = 203
    WINDOWS = 204
    UEFI = 205
    DOS = 206


class QL_OUTPUT(IntEnum):
    OFF = 1
    DEFAULT = 2
    DISASM = 3
    DEBUG = 4
    DUMP = 5


class QL_DEBUGGER(IntEnum):
    GDB = 1
    IDAPRO = 2
    QDB = 3


class QL_INTERCEPT(IntEnum):
    CALL = 1
    ENTER = 2
    EXIT = 3

QL_DEBUGGER_ALL = [QL_DEBUGGER.IDAPRO, QL_DEBUGGER.GDB, QL_DEBUGGER.QDB]
QL_ARCH_ALL     = [QL_ARCH.X86, QL_ARCH.X8664, QL_ARCH.ARM, QL_ARCH.ARM_THUMB, QL_ARCH.ARM64, QL_ARCH.MIPS, QL_ARCH.A8086]
QL_ARCH_16BIT   = [QL_ARCH.A8086]
QL_ARCH_32BIT   = [QL_ARCH.ARM, QL_ARCH.ARM_THUMB, QL_ARCH.MIPS, QL_ARCH.X86]
QL_ARCH_64BIT   = [QL_ARCH.ARM64, QL_ARCH.X8664] 
QL_ARCH_ENDIAN  = [QL_ARCH.MIPS, QL_ARCH.ARM]
QL_OS_ALL       = [QL_OS.LINUX, QL_OS.FREEBSD, QL_OS.MACOS, QL_OS.WINDOWS, QL_OS.UEFI, QL_OS.DOS]
QL_OS_NONPID    = [QL_OS.DOS, QL_OS.UEFI]
QL_OS_POSIX     = [QL_OS.LINUX, QL_OS.FREEBSD, QL_OS.MACOS]

QL_HOOK_BLOCK = 0b1
QL_CALL_BLOCK = 0b10

debugger_map = {
        "gdb": QL_DEBUGGER.GDB,
        "ida": QL_DEBUGGER.IDAPRO,
        "qdb": QL_DEBUGGER.QDB,
    }

arch_map = {
        "x86": QL_ARCH.X86,
        "x8664": QL_ARCH.X8664,
        "mips": QL_ARCH.MIPS,
        "arm": QL_ARCH.ARM,
        "arm_thumb": QL_ARCH.ARM_THUMB,
        "arm64": QL_ARCH.ARM64,
        "a8086": QL_ARCH.A8086,
    }

os_map = {
        "linux": QL_OS.LINUX,
        "macos": QL_OS.MACOS,
        "freebsd": QL_OS.FREEBSD,
        "windows": QL_OS.WINDOWS,
        "uefi": QL_OS.UEFI,
        "dos": QL_OS.DOS,
}


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

class FileReader:
    def __init__(self, binary):
        self.binary = binary
        self.offset = 0

    def read(self, size):
        data = self.binary[self.offset : self.offset + size]
        self.offset += size
        #print(data)
        return data

    def setOffset(self, offset):
        self.offset = offset

    def readString(self, align):
        str_size = 0
        while True:
            if self.binary[self.offset + str_size] != 0:
                str_size += 1
            else:
                str_size += 1
                break

        result = self.binary[self.offset : self.offset + str_size].decode("utf-8")

        if str_size % align != 0:
            self.offset += ((str_size // align) + 1) * align
        else:
            self.offset += str_size

        return result.strip(b'\x00'.decode())


def hexdump(src, length=16):
    FILTER = ''.join([(len(repr(chr(x))) == 3) and chr(x) or '.' for x in range(256)])
    lines = []
    for c in range(0, len(src), length):
        chars = src[c:c+length]
        hex = ' '.join(["%02x" % x for x in chars])
        printable = ''.join(["%s" % ((x <= 127 and FILTER[x]) or '.') for x in chars])
        lines.append("%04x  %-*s  %s\n" % (c, length*3, hex, printable))
    return ''.join(lines)



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct, ipaddress


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *


def ql_bin_to_ip(ip):
    return ipaddress.ip_address(ip).compressed


def ql_syscall_socket(ql, socket_domain, socket_type, socket_protocol, *args, **kw):
    if ql.archtype== QL_ARCH.MIPS and socket_type == 2:
        socket_type = 1
    elif ql.archtype== QL_ARCH.MIPS and socket_type == 1:
        socket_type = 1

    idx = -1
    for i in range(256):
        if ql.os.fd[i] == 0:
            idx = i
            break
    try:
        if idx == -1:
            regreturn = -1
        else:
            if ql.output == QL_OUTPUT.DEBUG: # set REUSEADDR options under debug mode
                ql.os.fd[idx] = ql_socket.open(socket_domain, socket_type, socket_protocol, (socket.SOL_SOCKET, socket.SO_REUSEADDR, 1))
            else:
                ql.os.fd[idx] = ql_socket.open(socket_domain, socket_type, socket_protocol)

            regreturn = (idx)
    except:
        regreturn = -1

    socket_type = socket_type_mapping(socket_type, ql.archtype)
    socket_domain = socket_domain_mapping(socket_domain, ql.archtype)
    ql.log.debug("socket(%s, %s, %s) = %d" % (socket_domain, socket_type, socket_protocol, regreturn))

    return regreturn


def ql_syscall_connect(ql, connect_sockfd, connect_addr, connect_addrlen, *args, **kw):
    AF_UNIX = 1
    AF_INET = 2
    sock_addr = ql.mem.read(connect_addr, connect_addrlen)
    family = ql.unpack16(sock_addr[ : 2])
    s = ql.os.fd[connect_sockfd]
    ip = b''
    sun_path = b''
    port = 0
    try:
        if s.family == family:
            if s.family == AF_UNIX:
                sun_path = sock_addr[2 : ].split(b"\x00")[0]
                sun_path = ql.os.transform_to_real_path(sun_path.decode())
                s.connect(sun_path)
                regreturn = 0
            elif s.family == AF_INET:
                port, host = struct.unpack(">HI", sock_addr[2:8])
                ip = ql_bin_to_ip(host)
                s.connect((ip, port))
                regreturn = 0
            else:
                regreturn = -1
        else:
            regreturn = -1
    except:
        regreturn = -1

    if s.family == AF_UNIX:
        ql.log.debug("connect(%s) = %d" % (sun_path, regreturn))
    elif s.family == AF_INET:
        ql.log.debug("connect(%s, %d) = %d" % (ip, port, regreturn))
    else:
        ql.log.debug("connect() = %d" % (regreturn))
    return regreturn


def ql_syscall_setsockopt(ql, *args, **kw):
    regreturn = 0
    return regreturn


def ql_syscall_shutdown(ql, shutdown_fd, shutdown_how, *args, **kw):
    if shutdown_fd >=0 and shutdown_fd < 256 and ql.os.fd[shutdown_fd] != 0:
        try:
            ql.os.fd[shutdown_fd].shutdown(shutdown_how)
            regreturn = 0
        except:
            regreturn = -1
    return regreturn


def ql_syscall_bind(ql, bind_fd, bind_addr, bind_addrlen,  *args, **kw):
    regreturn = 0

    if ql.archtype == QL_ARCH.X8664:
        data = ql.mem.read(bind_addr, 8)
    else:
        data = ql.mem.read(bind_addr, bind_addrlen)

    sin_family = ql.unpack16(data[:2]) or ql.os.fd[bind_fd].family
    port, host = struct.unpack(">HI", data[2:8])
    host = ql_bin_to_ip(host)

    if ql.root == False and port <= 1024:
        port = port + 8000

    if sin_family == 1:
        path = data[2 : ].split(b'\x00')[0]
        path = ql.os.transform_to_real_path(path.decode())
        ql.log.info(path)
        ql.os.fd[bind_fd].bind(path)

    # need a proper fix, for now ipv4 comes first
    elif sin_family == 2 and ql.os.bindtolocalhost == True:
        ql.os.fd[bind_fd].bind(('127.0.0.1', port))
        host = "127.0.0.1"

    # IPv4 should comes first
    elif ql.os.ipv6 == True and sin_family == 10 and ql.os.bindtolocalhost == True:
        ql.os.fd[bind_fd].bind(('::1', port))
        host = "::1"

    elif ql.os.bindtolocalhost == False:
        ql.os.fd[bind_fd].bind((host, port))

    else:
        regreturn = -1

    if ql.code:
        regreturn = 0

    if sin_family == 1:
        ql.log.debug("bind(%d, %s, %d) = %d" % (bind_fd, path, bind_addrlen, regreturn))
    else:
        ql.log.debug("bind(%d,%s:%d,%d) = %d" % (bind_fd, host, port, bind_addrlen,regreturn))
        ql.log.debug("syscall bind host: %s and port: %i sin_family: %i" % (ql_bin_to_ip(host), port, sin_family))

    return regreturn


def ql_syscall_getsockname(ql, sockfd, addr, addrlenptr, *args, **kw):
    if sockfd < 256 and ql.os.fd[sockfd] != 0:
        host, port = ql.os.fd[sockfd].getsockname()
        data = struct.pack("<h", int(ql.os.fd[sockfd].family))
        data += struct.pack(">H", port)
        data += ipaddress.ip_address(host).packed
        addrlen = ql.mem.read(addrlenptr, 4)
        addrlen = ql.unpack32(addrlen)
        data = data[:addrlen]
        ql.mem.write(addr, data)
        regreturn = 0
    else:
        regreturn = -1

    ql.log.debug("getsockname(%d, 0x%x, 0x%x) = %d" % (sockfd, addr, addrlenptr, regreturn))
    return regreturn  


def ql_syscall_getpeername(ql, sockfd, addr, addrlenptr, *args, **kw):
    if sockfd < 256 and ql.os.fd[sockfd] != 0:
        host, port = ql.os.fd[sockfd].getpeername()
        data = struct.pack("<h", int(ql.os.fd[sockfd].family))
        data += struct.pack(">H", port)
        data += ipaddress.ip_address(host).packed
        addrlen = ql.mem.read(addrlenptr, 4)
        addrlen = ql.unpack32(addrlen)
        data = data[:addrlen]
        ql.mem.write(addr, data)
        regreturn = 0
    else:
        regreturn = -1

    ql.log.debug("getpeername(%d, 0x%x, 0x%x) = %d" % (sockfd, addr, addrlenptr, regreturn))
    return regreturn  


def ql_syscall_listen(ql, listen_sockfd, listen_backlog, *args, **kw):
    if listen_sockfd < 256 and ql.os.fd[listen_sockfd] != 0:
        try:
            ql.os.fd[listen_sockfd].listen(listen_backlog)
            regreturn = 0
        except:
            if ql.output == QL_OUTPUT.DEBUG:
                raise
            regreturn = -1
    else:
        regreturn = -1
    return regreturn


def ql_syscall_accept(ql, accept_sockfd, accept_addr, accept_addrlen, *args, **kw):
    def inet_addr(ip):
        ret = b''
        tmp = ip.split('.')
        if len(tmp) != 4:
            return ret
        for i in tmp[ : : -1]:
            ret += bytes([int(i)])
        return ret
    try:
        conn, address = ql.os.fd[accept_sockfd].accept()
        if conn == None:
            return -1

        idx = -1
        for i in range(256):
            if ql.os.fd[i] == 0:
                idx = i
                break
        if idx == -1:
            regreturn = -1
        else:
            ql.os.fd[idx] = conn
            regreturn = idx

        if ql.code == None and accept_addr !=0 and accept_addrlen != 0:
            tmp_buf = ql.pack16(conn.family)
            tmp_buf += ql.pack16(address[1])
            tmp_buf += inet_addr(address[0])
            tmp_buf += b'\x00' * 8
            ql.mem.write(accept_addr, tmp_buf)
            ql.mem.write(accept_addrlen, ql.pack32(16))
    except:
        if ql.output in (QL_OUTPUT.DEBUG, QL_OUTPUT.DUMP):
            raise
        regreturn = -1

    return regreturn


def ql_syscall_recv(ql, recv_sockfd, recv_buf, recv_len, recv_flags, *args, **kw):
    if recv_sockfd < 256 and ql.os.fd[recv_sockfd] != 0:
        tmp_buf = ql.os.fd[recv_sockfd].recv(recv_len, recv_flags)
        if tmp_buf:
            ql.log.debug("recv() CONTENT:")
            ql.log.debug("%s" % tmp_buf)
        ql.mem.write(recv_buf, tmp_buf)
        regreturn = len(tmp_buf)
    else:
        regreturn = -1
    return regreturn


def ql_syscall_send(ql, send_sockfd, send_buf, send_len, send_flags, *args, **kw):
    regreturn = 0
    if send_sockfd < 256 and ql.os.fd[send_sockfd] != 0:
        try:
            ql.log.debug("debug send() start")
            tmp_buf = ql.mem.read(send_buf, send_len)  
            ql.log.debug("fd is " + str(send_sockfd))
            ql.log.debug("send() CONTENT:")
            ql.log.debug("%s" % str(tmp_buf))
            ql.log.debug("send() flag is " + str(send_flags))
            ql.log.debug("send() len is " + str(send_len))
            regreturn = ql.os.fd[send_sockfd].send(bytes(tmp_buf), send_flags)
            ql.log.debug("debug send end")
        except:
            ql.log.info(sys.exc_info()[0])
            if ql.output in (QL_OUTPUT.DEBUG, QL_OUTPUT.DUMP):
                raise
    else:
        regreturn = -1
    return regreturn


def ql_syscall_recvfrom(ql, recvfrom_sockfd, recvfrom_buf, recvfrom_len, recvfrom_flags, recvfrom_addr, recvfrom_addrlen, *args, **kw):
    # For x8664, recvfrom() is called finally when calling recv() in TCP communications
    SOCK_STREAM = 1
    if ql.os.fd[recvfrom_sockfd].socktype == SOCK_STREAM:
        return ql_syscall_recv(ql, recvfrom_sockfd, recvfrom_buf, recvfrom_len, recvfrom_flags, *args, **kw)
    else:
        if recvfrom_sockfd < 256 and ql.os.fd[recvfrom_sockfd] != 0:
            tmp_buf, tmp_addr = ql.os.fd[recvfrom_sockfd].recvfrom(recvfrom_len, recvfrom_flags)
            if tmp_buf:
                ql.log.debug("recvfrom() CONTENT:")
                ql.log.debug("%s" % tmp_buf)

            sin_family = int(ql.os.fd[recvfrom_sockfd].family)
            data = struct.pack("<h", sin_family)
            if sin_family == 1:
                ql.log.debug("recvfrom() path is " + tmp_addr)
                data += tmp_addr.encode()
            else:
                ql.log.debug("recvfrom() addr is %s:%d" % (tmp_addr[0], tmp_addr[1]))
                data += struct.pack(">H", tmp_addr[1])
                data += ipaddress.ip_address(tmp_addr[0]).packed
                addrlen = ql.unpack(ql.mem.read(recvfrom_addrlen, ql.pointersize))
                data = data[:addrlen]
            ql.mem.write(recvfrom_addr, data)

            ql.mem.write(recvfrom_buf, tmp_buf)
            regreturn = len(tmp_buf)
        else:
            regreturn = -1

        return regreturn


def ql_syscall_sendto(ql, sendto_sockfd, sendto_buf, sendto_len, sendto_flags, sendto_addr, sendto_addrlen, *args, **kw):
    # For x8664, sendto() is called finally when calling send() in TCP communications
    SOCK_STREAM = 1
    if ql.os.fd[sendto_sockfd].socktype == SOCK_STREAM:
        return ql_syscall_send(ql, sendto_sockfd, sendto_buf, sendto_len, sendto_flags, *args, **kw)
    else:
        regreturn = 0
        if sendto_sockfd < 256 and ql.os.fd[sendto_sockfd] != 0:
            try:
                ql.log.debug("debug sendto() start")
                tmp_buf = ql.mem.read(sendto_buf, sendto_len)

                if ql.archtype== QL_ARCH.X8664:
                    data = ql.mem.read(sendto_addr, 8)
                else:
                    data = ql.mem.read(sendto_addr, sendto_addrlen)

                sin_family, = struct.unpack("<h", data[:2])
                port, host = struct.unpack(">HI", data[2:8])
                host = ql_bin_to_ip(host)

                if sin_family == 1:
                    path = data[2 : ].split(b'\x00')[0]
                    path = ql.os.transform_to_real_path(path.decode())

                ql.log.debug("fd is " + str(sendto_sockfd))
                ql.log.debug("sendto() CONTENT:")
                ql.log.debug("%s" % tmp_buf)
                ql.log.debug("sendto() flag is " + str(sendto_flags))
                ql.log.debug("sendto() len is " + str(sendto_len))
                if sin_family == 1:
                    ql.log.debug("sendto() path is " + str(path))
                    regreturn = ql.os.fd[sendto_sockfd].sendto(bytes(tmp_buf), sendto_flags, path)
                else:
                    ql.log.debug("sendto() addr is %s:%d" % (host, port))
                    regreturn = ql.os.fd[sendto_sockfd].sendto(bytes(tmp_buf), sendto_flags, (host, port))
                ql.log.debug("debug sendto end")
            except:
                ql.log.info(sys.exc_info()[0])
                if ql.output in (QL_OUTPUT.DEBUG, QL_OUTPUT.DUMP):
                    raise
        else:
            regreturn = -1

        return regreturn



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from binascii import crc32

from qiling.os.const import *
from qiling.os.uefi.const import *
from qiling.os.uefi.fncc import *
from qiling.os.uefi.utils import *
from qiling.os.uefi.ProcessorBind import *
from qiling.os.uefi.UefiSpec import *
from qiling.os.uefi.protocols import common

# TODO: find a better solution than hardcoding this
pointer_size = 8

@dxeapi(params = {
	"NewTpl" : ULONGLONG		# EFI_TPL
})
def hook_RaiseTPL(ql, address, params):
	prev_tpl = ql.loader.tpl
	ql.loader.tpl = params["NewTpl"]

	return prev_tpl

@dxeapi(params = {
	"OldTpl": ULONGLONG			# EFI_TPL
})
def hook_RestoreTPL(ql, address, params):
	ql.loader.tpl = params["OldTpl"]

@dxeapi(params = {
	"type"		: INT,			# EFI_ALLOCATE_TYPE
	"MemoryType": INT,			# EFI_MEMORY_TYPE
	"Pages"		: ULONGLONG,	# UINTN
	"Memory"	: POINTER		# PTR(EFI_PHYSICAL_ADDRESS)
})
def hook_AllocatePages(ql, address, params):
	alloc_size = params["Pages"] * PAGE_SIZE

	if params['type'] == EFI_ALLOCATE_TYPE.AllocateAddress:
		address = read_int64(ql, params["Memory"])

		# TODO: check the range [address, address + alloc_size] is available first
		ql.mem.map(address, alloc_size)
	else:
		# TODO: allocate memory according to 'MemoryType'
		address = ql.loader.dxe_context.heap.alloc(alloc_size)

		if address == 0:
			return EFI_OUT_OF_RESOURCES

		write_int64(ql, params["Memory"], address)

	return EFI_SUCCESS

@dxeapi(params = {
	"Memory"	: ULONGLONG,	# EFI_PHYSICAL_ADDRESS
	"Pages"		: ULONGLONG		# UINTN
})
def hook_FreePages(ql, address, params):
	address = params["Memory"]
	alloc_size = params["Pages"] * PAGE_SIZE

	ret = ql.mem.free(address, alloc_size)

	return EFI_SUCCESS if ret else EFI_INVALID_PARAMETER

@dxeapi(params = {
	"MemoryMapSize"		: POINTER,	# PTR(UINTN)
	"MemoryMap"			: POINTER,	# PTR(EFI_MEMORY_DESCRIPTOR)
	"MapKey"			: POINTER,	# PTR(UINTN)
	"DescriptorSize"	: POINTER,	# PTR(UINTN)
	"DescriptorVersion"	: POINTER	# PTR(UINT32)
})
def hook_GetMemoryMap(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"PoolType"	: INT,		# EFI_MEMORY_TYPE
	"Size"		: INT,		# UINTN
	"Buffer"	: POINTER	# PTR(PTR(VOID))
})
def hook_AllocatePool(ql, address, params):
	# TODO: allocate memory acording to "PoolType"
	address = ql.loader.dxe_context.heap.alloc(params["Size"])
	write_int64(ql, params["Buffer"], address)

	return EFI_SUCCESS if address else EFI_OUT_OF_RESOURCES

@dxeapi(params = {
	"Buffer": POINTER # PTR(VOID)
})
def hook_FreePool(ql, address, params):
	address = params["Buffer"]
	ret = ql.loader.dxe_context.heap.free(address)

	return EFI_SUCCESS if ret else EFI_INVALID_PARAMETER

@dxeapi(params = {
	"Type"			: UINT,		# UINT32
	"NotifyTpl"		: UINT,		# EFI_TPL
	"NotifyFunction": POINTER,	# EFI_EVENT_NOTIFY
	"NotifyContext"	: POINTER,	# PTR(VOID)
	"Event"			: POINTER	# PTR(EFI_EVENT)
})
def hook_CreateEvent(ql, address, params):
	return CreateEvent(ql, params)

@dxeapi(params = {
	"Event"			: POINTER,		# EFI_EVENT
	"Type"			: ULONGLONG,	# EFI_TIMER_DELAY
	"TriggerTime"	: ULONGLONG		# UINT64
})
def hook_SetTimer(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"NumberOfEvents": ULONGLONG,	# UINTN
	"Event"			: POINTER,		# PTR(EFI_EVENT)
	"Index"			: POINTER,		# PTR(UINTN)
})
def hook_WaitForEvent(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"Event": POINTER # EFI_EVENT
})
def hook_SignalEvent(ql, address, params):
	event_id = params["Event"]

	if event_id in ql.loader.events:
		signal_event(ql, event_id)
		return EFI_SUCCESS
	else:
		return EFI_INVALID_PARAMETER

@dxeapi(params = {
	"Event": POINTER # EFI_EVENT
})
def hook_CloseEvent(ql, address, params):
	event_id = params["Event"]
	del ql.loader.events[event_id]

	return EFI_SUCCESS

@dxeapi(params = {
	"Event": POINTER # EFI_EVENT
})
def hook_CheckEvent(ql, address, params):
	event_id = params["Event"]

	return EFI_SUCCESS if ql.loader.events[event_id]["Set"] else EFI_NOT_READY

@dxeapi(params = {
	"Handle"		: POINTER,		# PTR(EFI_HANDLE)
	"Protocol"		: GUID,			# PTR(EFI_GUID)
	"InterfaceType"	: ULONGLONG,	# EFI_INTERFACE_TYPE
	"Interface"		: POINTER,		# PTR(VOID)
})
def hook_InstallProtocolInterface(ql, address, params):
	return common.InstallProtocolInterface(ql.loader.dxe_context, params)

@dxeapi(params = {
	"Handle"		: POINTER,	# EFI_HANDLE
	"Protocol"		: GUID,		# PTR(EFI_GUID)
	"OldInterface"	: POINTER,	# PTR(VOID)
	"NewInterface"	: POINTER	# PTR(VOID)
})
def hook_ReinstallProtocolInterface(ql, address, params):
	handle = params["Handle"]

	if handle not in ql.loader.dxe_context.protocols:
		return EFI_NOT_FOUND

	dic = ql.loader.dxe_context.protocols[handle]
	protocol = params["Protocol"]

	if protocol not in dic:
		return EFI_NOT_FOUND

	dic[protocol] = params["NewInterface"]

	return EFI_SUCCESS

@dxeapi(params = {
	"Handle"	: POINTER,	# EFI_HANDLE
	"Protocol"	: GUID,		# PTR(EFI_GUID)
	"Interface"	: POINTER	# PTR(VOID)
})
def hook_UninstallProtocolInterface(ql, address, params):
	return common.UninstallProtocolInterface(ql.loader.dxe_context, params)

@dxeapi(params = {
	"Handle"	: POINTER,	# EFI_HANDLE
	"Protocol"	: GUID,		# PTR(EFI_GUID)
	"Interface"	: POINTER	# PTR(PTR(VOID))
})
def hook_HandleProtocol(ql, address, params):
	return common.HandleProtocol(ql.loader.dxe_context, params)

@dxeapi(params = {
	"Protocol"		: GUID,		# PTR(EFI_GUID)
	"Event"			: POINTER,	# EFI_EVENT
	"Registration"	: POINTER	# PTR(PTR(VOID))
})
def hook_RegisterProtocolNotify(ql, address, params):
	event = params['Event']
	proto = params["Protocol"]

	if event in ql.loader.events:
		ql.loader.events[event]['Guid'] = proto
		check_and_notify_protocols(ql)

		return EFI_SUCCESS

	return EFI_INVALID_PARAMETER

@dxeapi(params = {
	"SearchType": INT,		# EFI_LOCATE_SEARCH_TYPE
	"Protocol"	: GUID,		# PTR(EFI_GUID)
	"SearchKey"	: POINTER,	# PTR(VOID)
	"BufferSize": POINTER,	# PTR(UINTN)
	"Buffer"	: POINTER	# PTR(EFI_HANDLE)
})
def hook_LocateHandle(ql, address, params):
	return common.LocateHandle(ql.loader.dxe_context, params)

@dxeapi(params = {
	"Protocol"	: GUID,		# PTR(EFI_GUID)
	"DevicePath": POINTER,	# PTR(PTR(EFI_DEVICE_PATH_PROTOCOL))
	"Device"	: POINTER	# PTR(EFI_HANDLE)
})
def hook_LocateDevicePath(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"Guid"	: GUID,		# PTR(EFI_GUID)
	"Table"	: POINTER	# PTR(VOID)
})
def hook_InstallConfigurationTable(ql, address, params):
	return common.InstallConfigurationTable(ql.loader.dxe_context, params)

@dxeapi(params = {
	"BootPolicy"		: BOOL,			# BOOLEAN
	"ParentImageHandle"	: POINTER,		# EFI_HANDLE
	"DevicePath"		: POINTER,		# PTR(EFI_DEVICE_PATH_PROTOCOL)
	"SourceBuffer"		: POINTER,		# PTR(VOID)
	"SourceSize"		: ULONGLONG,	# UINTN
	"ImageHandle"		: POINTER		# PTR(EFI_HANDLE)
})
def hook_LoadImage(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"ImageHandle"	: POINTER,	# EFI_HANDLE
	"ExitDataSize"	: POINTER,	# PTR(UINTN)
	"ExitData"		: POINTER	# PTR(PTR(CHAR16))
})
def hook_StartImage(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"ImageHandle"	: POINTER,		# EFI_HANDLE
	"ExitStatus"	: ULONGLONG,	# EFI_STATUS
	"ExitDataSize"	: ULONGLONG,	# UINTN
	"ExitData"		: POINTER		# PTR(CHAR16)
})
def hook_Exit(ql, address, params):
	ql.uc.emu_stop()

	return EFI_SUCCESS

@dxeapi(params = {
	"ImageHandle" : POINTER # EFI_HANDLE
})
def hook_UnloadImage(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"ImageHandle"	: POINTER,	# EFI_HANDLE
	"MapKey"		: ULONGLONG	# UINTN
})
def hook_ExitBootServices(ql, address, params):
	ql.uc.emu_stop()

	return EFI_SUCCESS

@dxeapi(params = {
	"Count": POINTER # PTR(UINT64)
})
def hook_GetNextMonotonicCount(ql, address, params):
	out = params["Count"]

	ql.os.monotonic_count += 1
	write_int64(ql, out, ql.os.monotonic_count)

	return EFI_SUCCESS

@dxeapi(params = {
	"Microseconds": ULONGLONG # UINTN
})
def hook_Stall(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"Timeout"		: ULONGLONG,	# UINTN
	"WatchdogCode"	: ULONGLONG,	# UINT64
	"DataSize"		: ULONGLONG,	# UINTN
	"WatchdogData"	: POINTER		# PTR(CHAR16)
})
def hook_SetWatchdogTimer(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"ControllerHandle"		: POINTER,	# EFI_HANDLE
	"DriverImageHandle"		: POINTER,	#PTR(EFI_HANDLE)
	"RemainingDevicePath"	: POINTER,	# PTR(EFI_DEVICE_PATH_PROTOCOL)
	"Recursive"				: BOOL		# BOOLEAN
})
def hook_ConnectController(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"ControllerHandle"	: POINTER,	# EFI_HANDLE
	"DriverImageHandle"	: POINTER,	# EFI_HANDLE
	"ChildHandle"		: POINTER	# EFI_HANDLE
})
def hook_DisconnectController(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"Handle"			: POINTER,	# EFI_HANDLE
	"Protocol"			: GUID,		# PTR(EFI_GUID)
	"Interface"			: POINTER,	# PTR(PTR(VOID))
	"AgentHandle"		: POINTER,	# EFI_HANDLE
	"ControllerHandle"	: POINTER,	# EFI_HANDLE
	"Attributes"		: UINT		# UINT32
})
def hook_OpenProtocol(ql, address, params):
	return common.LocateProtocol(ql.loader.dxe_context, params)

@dxeapi(params = {
	"Handle"			: POINTER,	# EFI_HANDLE
	"Protocol"			: GUID,		# PTR(EFI_GUID)
	"AgentHandle"		: POINTER,	# EFI_HANDLE
	"ControllerHandle"	: POINTER	# EFI_HANDLE
})
def hook_CloseProtocol(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"Handle"		: POINTER,	# EFI_HANDLE
	"Protocol"		: GUID,		# PTR(EFI_GUID)
	"EntryBuffer"	: POINTER,	# PTR(PTR(EFI_OPEN_PROTOCOL_INFORMATION_ENTRY))
	"EntryCount"	: POINTER	# PTR(UINTN)
})
def hook_OpenProtocolInformation(ql, address, params):
	return EFI_NOT_FOUND

@dxeapi(params = {
	"Handle"				: POINTER,	# EFI_HANDLE
	"ProtocolBuffer"		: POINTER,	# PTR(PTR(PTR(EFI_GUID)))
	"ProtocolBufferCount"	: POINTER	# PTR(UINTN)
})
def hook_ProtocolsPerHandle(ql, address, params):
	return EFI_SUCCESS

@dxeapi(params = {
	"SearchType": INT,		# EFI_LOCATE_SEARCH_TYPE
	"Protocol"	: GUID,		# PTR(EFI_GUID)
	"SearchKey"	: POINTER,	# PTR(VOID)
	"NoHandles"	: POINTER,	# PTR(UINTN)
	"Buffer"	: POINTER	# PTR(PTR(EFI_HANDLE))
})
def hook_LocateHandleBuffer(ql, address, params):
	buffer_size, handles = common.LocateHandles(ql.loader.dxe_context, params)
	write_int64(ql, params["NoHandles"], len(handles))

	if len(handles) == 0:
		return EFI_NOT_FOUND

	address = ql.loader.dxe_context.heap.alloc(buffer_size)
	write_int64(ql, params["Buffer"], address)

	if address == 0:
		return EFI_OUT_OF_RESOURCES

	for handle in handles:
		write_int64(ql, address, handle)
		address += pointer_size

	return EFI_SUCCESS

@dxeapi(params = {
	"Protocol"		: GUID,		# PTR(EFI_GUID)
	"Registration"	: POINTER,	# PTR(VOID)
	"Interface"		: POINTER	# PTR(PTR(VOID))
})
def hook_LocateProtocol(ql, address, params):
	return common.LocateProtocol(ql.loader.dxe_context, params)

@dxeapi(params = {
	"Handle" : POINTER # PTR(EFI_HANDLE)
	# ...
})
def hook_InstallMultipleProtocolInterfaces(ql, address, params):
	handle = read_int64(ql, params["Handle"])

	if handle == 0:
		handle = ql.loader.dxe_context.heap.alloc(pointer_size)

	dic = ql.loader.dxe_context.protocols.get(handle, {})

	# process elipsiss arguments
	index = 1
	while ql.os.get_param_by_index(index) != 0:
		GUID_ptr = ql.os.get_param_by_index(index)
		protocol_ptr = ql.os.get_param_by_index(index + 1)

		GUID = str(ql.os.read_guid(GUID_ptr))
		dic[GUID] = protocol_ptr

		ql.log.info(f' | {GUID} {protocol_ptr:#x}')
		index += 2

	ql.loader.dxe_context.protocols[handle] = dic
	check_and_notify_protocols(ql)
	write_int64(ql, params["Handle"], handle)

	return EFI_SUCCESS

@dxeapi(params = {
	"Handle" : POINTER # EFI_HANDLE
	# ...
})
def hook_UninstallMultipleProtocolInterfaces(ql, address, params):
	handle = params["Handle"]

	if handle not in ql.loader.dxe_context.protocols:
		return EFI_NOT_FOUND

	dic = ql.loader.dxe_context.protocols[handle]

	# process elipsiss arguments
	index = 1
	while ql.os.get_param_by_index(index) != 0:
		GUID_ptr = ql.os.get_param_by_index(index)
		protocol_ptr = ql.os.get_param_by_index(index + 1)

		GUID = str(ql.os.read_guid(GUID_ptr))

		if GUID not in dic:
			return EFI_INVALID_PARAMETER

		del dic[GUID]

		ql.log.info(f' | {GUID}, {protocol_ptr:#x}')
		index += 2

	return EFI_SUCCESS

@dxeapi(params = {
	"Data"		: POINTER,		# PTR(VOID)
	"DataSize"	: ULONGLONG,	# UINTN
	"Crc32"		: POINTER		# PTR(UINT32)
})
def hook_CalculateCrc32(ql, address, params):
	data = bytes(ql.mem.read(params['Data'], params['DataSize']))
	write_int32(ql, params['Crc32'], crc32(data))

	return EFI_SUCCESS

@dxeapi(params = {
	"Destination"	: POINTER,	# PTR(VOID)
	"Source"		: POINTER,	# PTR(VOID)
	"Length"		: SIZE_T	# UINTN
})
def hook_CopyMem(ql, address, params):
	data = bytes(ql.mem.read(params['Source'], params['Length']))
	ql.mem.write(params['Destination'], data)

@dxeapi(params = {
	"Buffer": POINTER,	# PTR(VOID)
	"Size"	: SIZE_T,	# UINTN
	"Value"	: BYTE		# UINT8
})
def hook_SetMem(ql, address, params):
	ptr = params["Buffer"]
	value = ql.pack8(params["Value"] & 0xff)

	# TODO: do this the Pythonic way
	for i in range(params["Size"]):
		ql.mem.write(ptr + i, value)

@dxeapi(params = {
	"Type"			: UINT,		# UINT32
	"NotifyTpl"		: ULONGLONG,# EFI_TPL
	"NotifyFunction": POINTER,	# EFI_EVENT_NOTIFY
	"NotifyContext"	: POINTER,	# PTR(VOID)
	"EventGroup"	: GUID,		# PTR(EFI_GUID)
	"Event"			: POINTER	# PTR(EFI_EVENT)
})
def hook_CreateEventEx(ql, address, params):
	return CreateEvent(ql, params)

def CreateEvent(ql, params):
	event_id = len(ql.loader.events)
	event_dic = {
		"NotifyFunction": params["NotifyFunction"],
		"NotifyContext"	: params["NotifyContext"],
		"Guid"			: "",
		"Set"			: False
	}

	if "EventGroup" in params:
		event_dic["EventGroup"] = params["EventGroup"]

	ql.loader.events[event_id] = event_dic
	write_int64(ql, params["Event"], event_id)

	return EFI_SUCCESS

def initialize(ql, gBS : int):
	descriptor = {
		'struct' : EFI_BOOT_SERVICES,
		'fields' : (
			('Hdr',							None),
			('RaiseTPL',					hook_RaiseTPL),
			('RestoreTPL',					hook_RestoreTPL),
			('AllocatePages',				hook_AllocatePages),
			('FreePages',					hook_FreePages),
			('GetMemoryMap',				hook_GetMemoryMap),
			('AllocatePool',				hook_AllocatePool),
			('FreePool',					hook_FreePool),
			('CreateEvent',					hook_CreateEvent),
			('SetTimer',					hook_SetTimer),
			('WaitForEvent',				hook_WaitForEvent),
			('SignalEvent',					hook_SignalEvent),
			('CloseEvent',					hook_CloseEvent),
			('CheckEvent',					hook_CheckEvent),
			('InstallProtocolInterface',	hook_InstallProtocolInterface),
			('ReinstallProtocolInterface',	hook_ReinstallProtocolInterface),
			('UninstallProtocolInterface',	hook_UninstallProtocolInterface),
			('HandleProtocol',				hook_HandleProtocol),
			('Reserved',					None),
			('RegisterProtocolNotify',		hook_RegisterProtocolNotify),
			('LocateHandle',				hook_LocateHandle),
			('LocateDevicePath',			hook_LocateDevicePath),
			('InstallConfigurationTable',	hook_InstallConfigurationTable),
			('LoadImage',					hook_LoadImage),
			('StartImage',					hook_StartImage),
			('Exit',						hook_Exit),
			('UnloadImage',					hook_UnloadImage),
			('ExitBootServices',			hook_ExitBootServices),
			('GetNextMonotonicCount',		hook_GetNextMonotonicCount),
			('Stall',						hook_Stall),
			('SetWatchdogTimer',			hook_SetWatchdogTimer),
			('ConnectController',			hook_ConnectController),
			('DisconnectController',		hook_DisconnectController),
			('OpenProtocol',				hook_OpenProtocol),
			('CloseProtocol',				hook_CloseProtocol),
			('OpenProtocolInformation',		hook_OpenProtocolInformation),
			('ProtocolsPerHandle',			hook_ProtocolsPerHandle),
			('LocateHandleBuffer',			hook_LocateHandleBuffer),
			('LocateProtocol',				hook_LocateProtocol),
			('InstallMultipleProtocolInterfaces',	hook_InstallMultipleProtocolInterfaces),
			('UninstallMultipleProtocolInterfaces',	hook_UninstallMultipleProtocolInterfaces),
			('CalculateCrc32',				hook_CalculateCrc32),
			('CopyMem',						hook_CopyMem),
			('SetMem',						hook_SetMem),
			('CreateEventEx',				hook_CreateEventEx)
		)
	}

	ql.os.monotonic_count = 0

	instance = init_struct(ql, gBS, descriptor)
	instance.saveTo(ql, gBS)

__all__ = [
	'initialize'
]


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

class QlRegisterManager():
    """
    This class exposes the ql.reg features that allows you to directly access
    or assign values to CPU registers of a particular architecture.

    Registers exposed are listed in the *_const.py files in the respective
    arch directories and are mapped to Unicorn Engine's definitions
    """
    def __init__(self, ql):
        self.register_mapping = {}
        self.reverse_mapping = {}
        self.ql = ql
        self.uc_pc = 0
        self.uc_sp = 0
        

    def __getattribute__(self, name):
        name = name.lower()
        if name in ("register_mapping", "ql", "uc_pc", "uc_sp"):
            return super(QlRegisterManager, self).__getattribute__(name)

        elif name in self.register_mapping:
            return self.ql.uc.reg_read(self.register_mapping[name])

        return super(QlRegisterManager, self).__getattribute__(name)


    def __setattr__(self, name, value):
        name=name.lower()
        if name in ("register_mapping", "ql", "uc_pc", "uc_sp"):
            super(QlRegisterManager, self).__setattr__(name, value)

        elif name in self.register_mapping:
            self.ql.uc.reg_write(self.register_mapping[name], value)
        else:
            super(QlRegisterManager, self).__setattr__(name, value)


    def expand_mapping(self, expanded_map):
        self.register_mapping = {**self.register_mapping, **expanded_map}


    # read register
    def read(self, register):
        if isinstance(register, str):
            register = self.register_mapping.get(register.lower(), None)
        return self.ql.uc.reg_read(register)


    def write(self, register, value):
        if isinstance(register, str):
            register = self.register_mapping.get(register.lower(), None) 
        return self.ql.uc.reg_write(register, value)


    def msr(self, msr, addr= None):
        if not addr:
            return self.ql.uc.msr_read(msr)
        else:
            self.ql.uc.msr_write(msr, addr)


    # ql.reg.save
    def save(self):
        reg_dict = {}
        for reg in self.register_mapping:
            reg_v = self.read(reg)
            reg_dict[reg] = reg_v
        return reg_dict


    # ql.reg.restore
    def restore(self, value = {}):
        for reg in self.register_mapping:
            reg_v= value[reg]
            self.write(reg, reg_v)


    # ql.reg.bit() - Register bit
    #FIXME: This needs to be implemented for all archs
    def bit(self, uc_reg):
        return self.ql.arch.get_reg_bit(uc_reg)


    # Generic methods to get SP and IP across Arch's #
    # These functions should only be used if the     #
    # caller is dealing with multiple Arch's         #
    def register_sp(self, sp_id):
        self.uc_sp = sp_id


    def register_pc(self, pc_id):
        self.uc_pc = pc_id


    @property
    def arch_pc(self):
        return self.ql.uc.reg_read(self.uc_pc)


    @arch_pc.setter
    def arch_pc(self, value):
        return self.ql.uc.reg_write(self.uc_pc, value)

    @property
    def arch_pc_name(self):
        return self.ql.reg.reverse_mapping[self.uc_pc]

    @property
    def arch_sp(self):
        return self.ql.uc.reg_read(self.uc_sp)


    @arch_sp.setter
    def arch_sp(self, value):
        return self.ql.uc.reg_write(self.uc_sp, value)


    def get_uc_reg(self, uc_reg_name):
        return self.register_mapping.get(uc_reg_name, None)


    def create_reverse_mapping(self):
        self.reverse_mapping = {v:k for k, v in self.register_mapping.items()}



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import ctypes, os, struct

from time import time


from qiling.os.macos.fncc import *
from qiling.os.macos.structs import *
from qiling.os.macos.utils import gen_stub_code
from qiling.os.macos.events.macos_structs import *
from qiling.os.macos.events.macos_policy import *


@macos_kernel_api(passthru=True, params={
    "scope": POINTER,
})
def hook__kauth_unlisten_scope(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "mbuf": POINTER,
})
def hook__mbuf_pkthdr_header(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "mbuf": POINTER,
})
def hook__mbuf_type(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "amount": INT,
    "address": POINTER,
})
def hook__OSAddAtomic(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "mbuf": POINTER,
})
def hook__mbuf_pkthdr_len(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "dst": POINTER,
    "src": STRING,
    "s": SIZE_T,
    "chk_size": SIZE_T,
})
def hook____strlcpy_chk(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "mask": INT,
    "address": POINTER,
})
def hook__OSBitOrAtomic(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "address": POINTER,
})
def hook__OSDecrementAtomic(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "cred": POINTER,
})
def hook__kauth_cred_getgid(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "cred": POINTER,
})
def hook__kauth_cred_getuid(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "tvp": POINTER,
})
def hook__microtime(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "address": POINTER,
})
def hook__OSIncrementAtomic(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "b": POINTER,
    "c": INT,
    "len": SIZE_T,
})
def hook__memset(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "s": POINTER,
    "n": SIZE_T,
    "format": STRING,
})
def hook__snprintf(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "fmt": STRING,
    "argp": POINTER,
    "putc": POINTER,
    "arg": POINTER,
    "radix": INT,
    "is_log": INT,
})
def hook____doprnt(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "ch": INT,
    "arg": POINTER,
})
def hook__snprintf_func(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "dst": POINTER,
    "src": POINTER,
    "s": SIZE_T,
    "chk_size": SIZE_T,
})
def hook____memcpy_chk(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "s1": STRING,
    "s2": STRING,
})
def hook__strcmp(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "s1": STRING,
    "s2": STRING,
})
def hook__strprefix(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "s1": STRING,
    "s2": STRING,
    "n": SIZE_T
})
def hook__strncmp(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "s": STRING,
})
def hook__strlen(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "protocol": INT,
    "address": POINTER,
    "string": POINTER,
    "len": SIZE_T,
})
def hook__inet_ntop(ql, address, params):
    return

@macos_kernel_api(params={
    "format": STRING,
})
def hook__IOLog(ql, address, params):
    return

@macos_kernel_api(params={
    "format": STRING,
})
def hook__printf(ql, address, params):
    return 0

@macos_kernel_api(passthru=True, params={
    "dst0": POINTER,
    "length": SIZE_T,
})
def hook__bzero(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "p": POINTER,
    "uap": POINTER,
})
def hook__sysctlbyname(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "oidp": POINTER,
    "arg1": POINTER,
    "arg2": INT,
    "req": POINTER
})
def hook__sysctl_handle_string(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "req": POINTER,
    "p": POINTER,
    "l": SIZE_T,
})
def hook__sysctl_old_user(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "req": POINTER,
    "p": POINTER,
    "l": SIZE_T,
})
def hook__sysctl_new_user(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "oidp": POINTER,
    "arg1": POINTER,
    "arg2": INT,
    "req": POINTER,
})
def hook__sysctl_handle_int(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "req": POINTER,
    "bigValue": ULONGLONG,
    "valueSize": SIZE_T,
    "pValue": POINTER,
    "changed": POINTER,
})
def hook__sysctl_io_number(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "vp": POINTER,
    "uio": POINTER,
    "alp": POINTER,
    "options": ULONGLONG,
    "vap": POINTER,
    "fndesc": POINTER,
    "ctx": POINTER,
})
def hook__vfs_attr_pack(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "alp": POINTER,
    "vap": POINTER,
    "obj_type": INT,
    "fixedsize": POINTER,
    "is_64bit": INT,
    "use_fork": INT,
})
def hook__getattrlist_setupvattr_all(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "vp": POINTER,
    "auio": POINTER,
    "alp": POINTER,
    "options": ULONGLONG,
    "vap": POINTER,
    "fndesc": POINTER,
    "ctx": POINTER,
    "is_bulk": INT,
    "vtype": INT,
    "fixedsize": SIZE_T,
})
def hook__vfs_attr_pack_internal(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "asp": POINTER,
    "vap": POINTER,
    "use_fork": INT,
})
def hook__getattrlist_fixupattrs(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "ab": POINTER,
    "source": STRING,
    "count": SIZE_T,
})
def hook__attrlist_pack_string(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "src": POINTER,
    "dst": POINTER,
    "len": SIZE_T,
})
def hook__bcopy(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "dst": POINTER,
    "src": POINTER,
    "n": SIZE_T,
})
def hook__memcpy(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "c_cp": POINTER,
    "n": INT,
    "uio": POINTER,
})
def hook__uiomove64(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "kernel_addr": POINTER,
    "user_addr": POINTER,
    "nbytes": SIZE_T,
})
def hook__copyout(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "a_uio": POINTER,
    "a_count": SIZE_T,
})
def hook__uio_update(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "user_addr": POINTER,
    "kernel_addr": POINTER,
    "nbytes": SIZE_T,
})
def hook__copyin(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "dst": POINTER,
    "src": POINTER,
    "s": SIZE_T,
    "chk_size": SIZE_T,
})
def hook____memmove_chk(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "mbuf": POINTER,
})
def hook__mbuf_len(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "mbuf": POINTER,
})
def hook__mbuf_data(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "provider": POINTER,
})
def hook___ZN9IOService5startEPS_(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "provider": POINTER, 
    "score": POINTER,
})
def hook___ZN9IOService5probeEPS_Pi(ql, address, params):
    return

@macos_kernel_api(params={
    "size": SIZE_T,
})
def hook___ZN8OSObjectnwEm(ql, address, params):
    size = params["size"]
    addr = ql.os.heap.alloc(size)
    return addr

@macos_kernel_api(params={
    "this": POINTER,
})
def hook___ZN9IOServiceC2EPK11OSMetaClass(ql, address, params):
    return 0;

@macos_kernel_api(params={
    "this": POINTER,
})
def hook___ZNK11OSMetaClass19instanceConstructedEv(ql, address, params):
    return 0;

@macos_kernel_api(params={
    "this": POINTER,
    "option": POINTER,
})
def hook___ZN9IOService4initEP12OSDictionary(ql, address, params):
    return 1

@macos_kernel_api(params={

})
def hook___ZN9IOService15registerServiceEj(ql, address, params):
    return 0

@macos_kernel_api(params={
    "provider": POINTER,
})
def hook___ZN9IOService6attachEPS_(ql, address, params):
    return 1

@macos_kernel_api(params={
    "provider": POINTER,
})
def hook___ZN9IOService6detachEPS_(ql, address, params):
    return 0

@macos_kernel_api(params={
    "this": POINTER,
})
def hook___ZN12IOUserClientC2EPK11OSMetaClass(ql, address, params):
    return 0

@macos_kernel_api(params={
    "this": POINTER,
    "owningTask": UINT,
    "securityID": POINTER,
    "type": UINT,
    "properties": POINTER,
})
def hook___ZN12IOUserClient12initWithTaskEP4taskPvjP12OSDictionary(ql, address, params):
    return 1

@macos_kernel_api(params={
    "str": STRING,
    "flags": UINT,
})
def hook__OSMalloc_Tagalloc(ql, address, params):
    s = params["str"]
    flags = params["flags"]

    OSMTag_addr = ql.os.heap.alloc(ctypes.sizeof(OSMallocTag))
    OSMTag = OSMallocTag(ql, OSMTag_addr)

    if flags & 0x1:
        OSMTag._OSMT_attr = 0x1
    OSMTag.OSMT_refcnt = 0x1
    OSMTag.OSMT_name = s[:64].encode()
    OSMTag.OSMT_state = 0xDEAB0000
    OSMTag.OSMT_link = queue_entry(POINTER64(0), POINTER64(0))
    OSMTag.updateToMem()
    return OSMTag_addr

def lck_grp_init(ql, grp, s, attr):
    grp.lck_grp_name = s[:64].encode()
    grp.lck_grp_refcnt = 1;
    if attr is not None:
        grp.lck_grp_attr = attr.grp_attr_val
    else:
        grp.lck_grp_attr = 0

    grp.updateToMem()

@macos_kernel_api(params={
    "str": STRING,
    "attr": POINTER,
})
def hook__lck_grp_alloc_init(ql, address, params):
    s = params["str"]
    attr = params["attr"]

    grp_addr = ql.os.heap.alloc(ctypes.sizeof(lck_grp_t))
    grp = lck_grp_t(ql, grp_addr)
    if params["attr"] > 0:
        attr = lck_grp_attr_t(ql, params["attr"])
        attr.loadFromMem()
    else:
        attr = None
    lck_grp_init(ql, grp, s, attr)
    return grp_addr

@macos_kernel_api(params={
    "grp": POINTER,
    "grp_name": STRING,
    "attr": POINTER
})
def hook__lck_grp_init(ql, address, params):
    grp_name = params["grp_name"]
    attr = params["attr"]
    
    if params["grp"] > 0:
        grp = lck_grp_t(ql, params["grp"])
        grp.loadFromMem()
    else:
        grp = None

    if params["attr"] > 0:
        attr = lck_grp_attr_t(ql, params["attr"])
        attr.loadFromMem()
    else:
        attr = None

    lck_grp_init(ql, grp, grp_name, attr)

@macos_kernel_api(params={
    "grp": POINTER,
})
def hook__lck_grp_free(ql, address, params):
    ql.os.heap.free(params["grp"])
    return

# Values are taken from i386 locks.h
def lck_mtx_ext_init(ql, lck, grp, attr):
    if (attr.lck_attr_val) & 0x1:
        lck.lck_mtx_deb.type = 0x4d4d;
        lck.lck_mtx_attr |= 0x1;

    lck.lck_mtx_grp = grp;

    if grp.lck_grp_attr & 0x1:
        lck.lck_mtx_attr |= 0x2;

    lck.lck_mtx.lck_mtx_is_ext = 1;
    lck.lck_mtx.lck_mtx_pad32 = 0xFFFFFFFF;
    lck.updateToMem()

def lck_mtx_init(ql, lck, grp, attr):
    lck_ext = None
    lck_attr = None

    if attr != 0:
        lck_attr_addr = attr
    else:
        lck_attr_addr = ql.kernel_extrn_symbols_detail[b"LockDefaultLckAttr"]["n_value"]

    lck_attr = lck_attr_t(ql, lck_attr_addr)
    if (lck_attr.lck_attr_val) & 0x1:
        lck_ext_addr = ql.os.heap.alloc(ctypes.sizeof(lck_mtx_ext_t))
        lck_ext = lck_mtx_ext_t(ql, lck_ext_addr)
        lck_mtx_ext_init(ql, lck_ext, grp, lck_attr);   
        lck.lck_mtx_tag = 0x07ff1007;
        lck.lck_mtx_ptr = lck_ext;
        
    else:
        lck.lck_mtx_owner = 0;
        lck.lck_mtx_state = 0;
    
    lck.lck_mtx_pad32 = 0xFFFFFFFF;
    lck.updateToMem()

def lck_mtx_init_ext(ql, lck, lck_ext, grp, attr):
    if attr != 0:
        lck_attr_addr = attr
    else:
        lck_attr_addr = ql.kernel_extrn_symbols_detail[b"LockDefaultLckAttr"]["n_value"]

    lck_attr = lck_attr_t(ql, lck_attr_addr)
    if (lck_attr.lck_attr_val) & 0x1:
        lck_mtx_ext_init(ql, lck_ext, grp, lck_attr);   
        lck.lck_mtx_tag = 0x07ff1007;
        lck.lck_mtx_ptr = lck_ext;
    else:
        lck.lck_mtx_owner = 0;
        lck.lck_mtx_state = 0;
    
    lck.lck_mtx_pad32 = 0xFFFFFFFF;
    lck.updateToMem()


@macos_kernel_api(params={
    "lck": POINTER,
    "lck_ext": POINTER,
    "grp": POINTER,
    "attr": POINTER,
})
def hook__lck_mtx_init_ext(ql, address, params):
    lck = lck_mtx_t(ql, params["lck"])
    lck_ext = lck_mtx_ext_t(ql, params["lck_ext"])
    grp = lck_grp_t(ql, params["grp"])
    attr = lck_attr_t(ql, params["attr"])
    lck_mtx_init_ext(ql, lck, lck_ext, grp, attr)
    return


@macos_kernel_api(params={
    "lck": POINTER,
    "grp": POINTER,
    "attr": POINTER
})
def hook__lck_mtx_init(ql, address, params):
    lck = lck_mtx_t(ql, params["lck"])
    grp = lck_grp_t(ql, params["grp"])
    attr = lck_attr_t(ql, params["attr"])
    lck_mtx_init(ql, lck, grp, attr)
    return

@macos_kernel_api(params={
    "grp": POINTER,
    "attr": POINTER,
})
def hook__lck_mtx_alloc_init(ql, address, params):
    lck_addr = ql.os.heap.alloc(ctypes.sizeof(lck_mtx_t))
    lck = lck_mtx_t(ql, lck_addr)
    if params["grp"] > 0:
        grp = lck_grp_t(ql, params["grp"])
        grp.loadFromMem()
    else:
        grp = None

    if params["attr"] > 0:
        attr = lck_attr_t(ql, params["attr"])
        attr.loadFromMem()
    else:
        attr = None
    lck_mtx_init(ql, lck, grp, attr)
    return lck_addr

@macos_kernel_api(params={
    "lock": POINTER,
})
def hook__lck_mtx_lock(ql, address, params):
    return 1

@macos_kernel_api(params={
    "lock": POINTER,
})
def hook__lck_mtx_unlock(ql, address, params):
    return 1

@macos_kernel_api(params={

})
def hook__proc_list_lock(ql, address, params):
    return

@macos_kernel_api(params={

})
def hook__proc_list_unlock(ql, address, params):
    return

@macos_kernel_api(params={
    "size": SIZE_T,
    "tag": POINTER,
})
def hook__OSMalloc(ql, address, params):
    size = params["size"]
    addr = ql.os.heap.alloc(size)
    return addr

@macos_kernel_api(params={
    "addr": POINTER,
    "size": UINT,
    "tag": POINTER,
})
def hook__OSFree(ql, address, params):
    ql.os.heap.free(params["addr"])
    return


@macos_kernel_api(params={
    "psize": POINTER,
    "canblock": BOOL,
    "site": POINTER,
})
def hook__kalloc_canblock(ql, address, params):
    size, = struct.unpack("<Q", ql.mem.read(params["psize"], 8))
    addr = ql.os.heap.alloc(size)
    return addr

@macos_kernel_api(params={
    "data": POINTER,
    "size": SIZE_T,
})
def hook__kfree(ql, address, params):
    ql.os.heap.free(params["data"])
    return

@macos_kernel_api(params={
    "size": SIZE_T,
    "type": INT,
    "flags": INT,
})
def hook___MALLOC(ql, address, params):
    size = params["size"]
    addr = ql.os.heap.alloc(size)
    return addr

@macos_kernel_api(params={
    "addr": POINTER,
    "type": INT,
})
def hook___FREE(ql, address, params):
    ql.os.heap.free(params["addr"])
    return

@macos_kernel_api(params={
    "lock": POINTER
})
def hook__lck_rw_lock_exclusive(ql, address, params):
    return 1

@macos_kernel_api(params={
    "lock": POINTER
})
def hook__lck_rw_unlock_exclusive(ql, address, params):
    return 1

@macos_kernel_api(params={
    "lock": POINTER
})
def hook__lck_rw_done(ql, address, params):
    return 2

@macos_kernel_api(params={
    "oldValue": POINTER,
    "newValue": POINTER,
    "address": POINTER
})
def hook__OSCompareAndSwap(ql, address, params):
    addr = params["address"]
    oldValue = params["oldValue"]
    newValue = params["newValue"]
    cur_val, = struct.unpack("<I", ql.mem.read(addr, 4))
    if cur_val == oldValue:
        ql.mem.write(addr, struct.pack("<Q", newValue))
        return 1
    return 0

@macos_kernel_api(params={
    "lock": POINTER,
    "type": UINT
})
def hook__lck_mtx_assert(ql, address, params):
    return

@macos_kernel_api(passthru=True, params={
    "format": STRING,
})
def hook__panic(ql, address, params):
    return

@macos_kernel_api(params={
    "tv_sec": POINTER,
    "tv_usec": POINTER,
})
def hook__clock_get_calendar_microtime(ql, address, params):
    now = time()
    ql.mem.write(params["tv_sec"], struct.pack("<Q", int(now)))
    ql.mem.write(params["tv_usec"], struct.pack("<Q", int(round(now * 1000))))
    return

@macos_kernel_api(params={
    "buf": POINTER,
    "size": UINT
})
def hook__read_random(ql, address, params):
    ql.mem.write(params["buf"], struct.pack("<I", 0xdeadbeaf))
    return

@macos_kernel_api(params={
    "buf": POINTER,
    "size": UINT,
})
def hook__read_erandom(ql, address, params):
    ql.mem.write(params["buf"], struct.pack("<I", 0xdeadbeaf))
    return

COPYIN = 0      # from user virtual to kernel virtual
COPYOUT = 1     # from kernel virtual to user virtual
COPYINSTR = 2   # string variant of copyout
COPYINPHYS = 3  # from user virtual to kernel physical
COPYOUTPHYS = 4 # from kernel physical to user virtual
COPYINWORD = 5  # from user virtual to kernel virtual

@macos_kernel_api(params={
    "copy_type": INT,
    "user_addr": POINTER,
    "kernel_addr": POINTER,
    "nbytes": SIZE_T,
    "lencopied": POINTER,
    "use_kernel_map": INT
})
def hook__copyio(ql, address, params):
    copy_type = params["copy_type"]
    user_addr = params["user_addr"]
    kernel_addr = params["kernel_addr"]
    nbytes = params["nbytes"]

    # print("User: 0x%x --- kernel: (%s) 0x%x --- Size: %d" % (user_addr, type(kernel_addr), kernel_addr, nbytes))
    if copy_type in (COPYIN, COPYINSTR, COPYINPHYS, COPYINWORD):
        data = ql.mem.read(user_addr, nbytes)
        ql.mem.write(kernel_addr, bytes(data))
    else:
        data = ql.mem.read(kernel_addr, nbytes)
        ql.mem.write(user_addr, bytes(data))
    return 0

@macos_kernel_api(params={
    "addr": POINTER,
})
def hook__kfree_addr(ql, address, params):
    ql.os.heap.free(params["addr"])
    return

@macos_kernel_api(params={
    "addr": POINTER,
})
def hook__fuulong(ql, address, params):
    result = ql.mem.read(params["addr"], 8)
    return struct.unpack("<Q", result)[0]

@macos_kernel_api(params={
    "addr": POINTER,
    "uword": ULONGLONG,
})
def hook__suulong(ql, address, params):
    ql.mem.write(params["addr"], struct.pack("<Q", params["uword"]))
    return 0

@macos_kernel_api(params={

})
def hook__proc_selfpid(ql, address, params):
    return 0x1337

@macos_kernel_api(params={
    "pid": INT,
    "buf": POINTER,
    "size": INT,
})
def hook__proc_name(ql, address, params):
    # FIXME Why p_comm is only 50, but MAXPATHLEN is 1024??
    proc = ql.os.ev_manager.proc_find(params["pid"])

    if proc is not None:
        ql.mem.write(params["buf"], proc.p_comm)
    return

@macos_kernel_api(params={

})
def hook__proc_selfppid(ql, address, params):
    return 0x13371337

@macos_kernel_api(params={
    "pid": INT,
})
def hook__proc_find(ql, address, params):
    proc = ql.os.ev_manager.proc_find(params["pid"])
    if proc is not None:
        return proc.base
    return 0

@macos_kernel_api(params={
    "ctx": POINTER,
})
def hook__vfs_context_proc(ql, address, params):
    return params["ctx"]

@macos_kernel_api(params={
    "cred": POINTER,
    "uid": INT,
    "gid": INT
})
def hook__kauth_cred_setuidgid(ql, address, params):
    temp_pcred_addr = ql.os.heap.alloc(ctypes.sizeof(ucred_t))
    temp_pcred = ucred_t(ql, temp_pcred_addr)

    temp_pcred.cr_uid = params["uid"];
    temp_pcred.cr_ruid = params["uid"];
    temp_pcred.cr_svuid = params["uid"];
    # temp_pcred->cr_flags = pcred->cr_flags;
    # if (pcred->cr_flags & CRF_NOMEMBERD) {
    #     temp_pcred->cr_gmuid = KAUTH_UID_NONE;
    #     temp_pcred->cr_flags |= CRF_NOMEMBERD;
    # } else {
    #     temp_pcred->cr_gmuid = uid;
    #     temp_pcred->cr_flags &= ~CRF_NOMEMBERD;
    # }
    temp_pcred.cr_ngroups = 1;
    # /* displacing a supplementary group opts us out of memberd */
    # if (kauth_cred_change_egid(&temp_cred, gid)) {
    #     temp_pcred->cr_gmuid = KAUTH_UID_NONE;
    #     temp_pcred->cr_flags |= CRF_NOMEMBERD;
    # }
    temp_pcred.cr_rgid = params["gid"];
    temp_pcred.cr_svgid = params["gid"];
    # temp_cred.cr_label = cred->cr_label;
    temp_pcred.updateToMem()
    return temp_pcred_addr

@macos_kernel_api(params={
    "vp": POINTER,
    "pathbuf": POINTER,
    "len": POINTER,
})
def hook__vn_getpath(ql, address, params):
    vp = vnode_t(ql, params["vp"])
    vp = vp.loadFromMem()
    name = ql.mem.string(vp.v_name.value).encode()
    ql.mem.write(params["pathbuf"], name)
    ql.mem.write(params["len"], struct.pack("<I", len(name)))
    return 0

@macos_kernel_api(params={
    "continuation": POINTER,
    "params": POINTER,
    "new_thread": POINTER,
})
def hook__kernel_thread_start(ql, address, params):
    return 0

@macos_kernel_api(params={

})
def hook__kauth_getuid(ql, address, params):
    return ql.os.ev_manager.cred.cr_posix.cr_uid

@macos_kernel_api(params={

})
def hook__kauth_getgid(ql, address, params):
    return ql.os.ev_manager.cred.cr_posix.cr_rgid

@macos_kernel_api(params={

})
def hook__msleep(ql, address, params):
    return 0

@macos_kernel_api(params={
    "so": POINTER,
    "sockname": POINTER,
    "socknamelen": INT,
})
def hook__sock_getsockname(ql, address, params):
    real_so = ql.os.ev_manager.sockets[params["so"]]
    ql.mem.write(params["sockname"], bytes(real_so)[:params["socknamelen"]])
    return 0

@macos_kernel_api(params={
    "cmd": STRING,
})
def hook__KUNCExecute(ql, address, params):
    # ql.log.debug("Starting userspace process: %s" % params["cmd"])
    return 0

######################## Custom events / callbacks ########################

@macos_kernel_api(params={
    "oidp": POINTER
})
def hook__sysctl_register_oid(ql, address, params):
    oidp = sysctl_oid_t(ql, params["oidp"])
    oidp = oidp.loadFromMem()

    oid_name = ql.mem.string(oidp.oid_name.value)
    oid_parent = b""
    for symname, symb in ql.loader.kernel_extrn_symbols_detail.items():
        if symb["n_value"] == oidp.oid_parent.value:
            st = symname.find(b"sysctl_") + len("sysctl_")
            en = symname.find(b"_children", st)
            oid_parent = symname[st:en]
            break

    if len(oid_parent) == 0:
        # print("/!\\ Get symbol from local kext")
        for symname, symb in ql.loader.kext_local_symbols.items():
            # print("\tComparing for %s: 0x%x and 0x%x" %(symname, symb["n_value"] + ql.loader.loadbase, oidp.oid_parent.value))
            if symb["n_value"] + ql.loader.loadbase == oidp.oid_parent.value:
                st = symname.find(b"sysctl_") + len("sysctl_")
                en = symname.find(b"_children", st)
                oid_parent = symname[st:en]
                break

    if len(oid_parent) == 0:
        # print("/!\\ Get symbol from external kext")
        for symname, symb in ql.loader.kext_extern_symbols.items():
            # print("\tComparing for %s: 0x%x and 0x%x" %(symname, symb["n_value"] + ql.loader.loadbase, oidp.oid_parent.value))
            if symb["n_value"] + ql.loader.loadbase == oidp.oid_parent.value:
                st = symname.find(b"sysctl_") + len("sysctl_")
                en = symname.find(b"_children", st)
                oid_parent = symname[st:en]
                break


    true_name = oid_parent.lstrip(b"_") + b"." + oid_name.encode() + b"\x00"
    if oidp.oid_handler.value != 0:
        ql.log.debug("New sysctl callback has been registered: %s" % true_name)

        ql.os.ev_manager.register(oidp.oid_handler.value, true_name, MacOSEventType.EV_SYSCTL, ev_obj=oidp, idx=0)
    return

@macos_kernel_api(params={
    "oidp": POINTER
})
def hook__sysctl_unregister_oid(ql, address, params):
    oidp = sysctl_oid_t(ql, params["oidp"])
    oidp = oidp.loadFromMem()

    oid_name = ql.mem.string(oidp.oid_name.value)
    oid_parent = b""
    for symname, symb in ql.loader.kernel_extrn_symbols_detail.items():
        if symb["n_value"] == oidp.oid_parent.value:
            st = symname.find(b"sysctl_") + len("sysctl_")
            en = symname.find(b"_children", st)
            oid_parent = symname[st:en]
            break

    true_name = oid_parent.lstrip(b"_") + b"." + oid_name.encode() + b"\x00"
    ql.log.debug("A sysctl event has been deregistered: %s" % true_name)
    ql.os.ev_manager.deregister(true_name)

@macos_kernel_api(params={
    "from_kernel": BOOL,
    "string_is_canonical": BOOL,
    "namestring": POINTER,
    "namestringlen": SIZE_T,
    "name": POINTER,
    "namelen": INT,
    "req": POINTER,
})
def hook__sysctl_root(ql, address, params):
    if params["string_is_canonical"] == "True":
        ev_name = ql.mem.read(params["namestring"], params["namestringlen"]).decode()
        ev_type = MacOSEventType.EV_SYSCTL
        event = ql.os.ev_manager.get_event_by_name_and_type(ev_name, ev_type)
        if event is not None:
            obj = event.event
            ql.os.ev_manager.emit(ev_name, ev_type, [obj.oid_arg1.value, obj.oid_arg2, params["req"]])
        else:
            ql.log.debug("Event not found (%s, %s)" % (ev_name, ev_type.name))
    return 0

@macos_kernel_api(params={
    "userctl": POINTER,
    "ctlref": POINTER,
})
def hook__ctl_register(ql, address, params):
    userctl = kern_ctl_reg_t(ql, params["userctl"])
    userctl = userctl.loadFromMem()
    ctl_name = ql.mem.string(params["userctl"]).encode()
    flag = 0
    if userctl.ctl_connect.value != 0:
        flag |= 1 << 0
        ql.os.ev_manager.register(userctl.ctl_connect.value, ctl_name, MacOSEventType.EV_CTL_CONNECT, ev_obj=userctl, idx=0)
    if userctl.ctl_disconnect.value != 0:
        flag |= 1 << 1
        ql.os.ev_manager.register(userctl.ctl_disconnect.value, ctl_name, MacOSEventType.EV_CTL_DISCONNECT, ev_obj=userctl, idx=0)
    if userctl.ctl_send.value != 0:
        flag |= 1 << 2
        ql.os.ev_manager.register(userctl.ctl_send.value, ctl_name, MacOSEventType.EV_CTL_SEND, ev_obj=userctl, idx=0)
    if userctl.ctl_setopt.value != 0:
        flag |= 1 << 3
        ql.os.ev_manager.register(userctl.ctl_setopt.value, ctl_name, MacOSEventType.EV_CTL_SETOPT, ev_obj=userctl, idx=0)
    if userctl.ctl_getopt.value != 0:
        flag |= 1 << 4
        ql.os.ev_manager.register(userctl.ctl_getopt.value, ctl_name, MacOSEventType.EV_CTL_GETOPT, ev_obj=userctl, idx=0)
    if userctl.ctl_rcvd.value != 0:
        flag |= 1 << 5
        ql.os.ev_manager.register(userctl.ctl_rcvd.value, ctl_name, MacOSEventType.EV_CTL_RCVD_FUNC, ev_obj=userctl, idx=0)
    if userctl.ctl_send_list.value != 0:
        flag |= 1 << 6
        ql.os.ev_manager.register(userctl.ctl_send_list.value, ctl_name, MacOSEventType.EV_CTL_SEND_LIST_FUNC, ev_obj=userctl, idx=0)
    if userctl.ctl_bind.value != 0:
        flag |= 1 << 7
        ql.os.ev_manager.register(userctl.ctl_bind.value, ctl_name, MacOSEventType.EV_CTL_BIND_FUNC, ev_obj=userctl, idx=0)


    ql.log.debug("New ctl callbacks has been registered: %s (%d/%d/%d/%d/%d/%d/%d/%d)" % (
        ctl_name, 
        (flag & (1 << 0)) != 0,
        (flag & (1 << 1)) != 0,
        (flag & (1 << 2)) != 0,
        (flag & (1 << 3)) != 0,
        (flag & (1 << 4)) != 0,
        (flag & (1 << 5)) != 0,
        (flag & (1 << 6)) != 0,
        (flag & (1 << 7)) != 0))

    ql.mem.write(params["ctlref"], struct.pack("<Q", userctl.base))
    return 0

@macos_kernel_api(params={
    "ctlref": POINTER
})
def hook__ctl_deregister(ql, address, params):
    userctl = kern_ctl_reg_t(ql, params["userctl"])
    userctl = userctl.loadFromMem()
    ctl_name = ql.mem.string(params["userctl"]).encode()
    ql.log.debug("A ctl event has been deregistered: %s" % ctl_name)
    ql.os.ev_manager.deregister(ctl_name)

@macos_kernel_api(params={
    "filter": POINTER,
    "domain": INT,
    "type": INT,
    "protocol": INT
})
def hook__sflt_register(ql, address, params):
    sf = sflt_filter_t(ql, params["filter"])
    sf = sf.loadFromMem()
    true_name = str(sf.sf_handle).encode()
    true_name += b"_" + ql.mem.string(sf.sf_name.value).encode()
    true_name += b"_" + str(params["domain"]).encode()
    true_name += b"_" + str(params["type"]).encode()
    true_name += b"_" + str(params["protocol"]).encode()
    flag = 0

    if sf.sf_unregistered != 0:
        flag |= 1 << 0
        ql.os.ev_manager.register(sf.sf_unregistered.value, true_name, MacOSEventType.EV_SFLT_UNREGISTERED, protocol=params["protocol"])
    if sf.sf_attach != 0:
        flag |= 1 << 1
        ql.os.ev_manager.register(sf.sf_attach.value, true_name, MacOSEventType.EV_SFLT_ATTACH, protocol=params["protocol"])
    if sf.sf_detach != 0:
        flag |= 1 << 2
        ql.os.ev_manager.register(sf.sf_detach.value, true_name, MacOSEventType.EV_SFLT_DETACH, protocol=params["protocol"])
    if sf.sf_notify != 0:
        flag |= 1 << 3
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_CONNECTING, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_CONNECTED, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_DISCONNECTING, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_DISCONNECTED, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_FLUSH_READ, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_SHUTDOWN, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_CANTRECVMORE, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_CANTSENDMORE, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_CLOSING, protocol=params["protocol"])
        ql.os.ev_manager.register(sf.sf_notify.value, true_name, MacOSEventType.EV_SFLT_NOTIFY_BOUND, protocol=params["protocol"])
    if sf.sf_getpeername != 0:
        flag |= 1 << 4
        ql.os.ev_manager.register(sf.sf_getpeername.value, true_name, MacOSEventType.EV_SFLT_GETPEERNAME, protocol=params["protocol"])
    if sf.sf_getsockname != 0:
        flag |= 1 << 5
        ql.os.ev_manager.register(sf.sf_getsockname.value, true_name, MacOSEventType.EV_SFLT_GETSOCKNAME, protocol=params["protocol"])
    if sf.sf_data_in != 0:
        flag |= 1 << 6
        ql.os.ev_manager.register(sf.sf_data_in.value, true_name, MacOSEventType.EV_SFLT_DATA_IN, protocol=params["protocol"])
    if sf.sf_data_out != 0:
        flag |= 1 << 7
        ql.os.ev_manager.register(sf.sf_data_out.value, true_name, MacOSEventType.EV_SFLT_DATA_OUT, protocol=params["protocol"])
    if sf.sf_connect_in != 0:
        flag |= 1 << 8
        ql.os.ev_manager.register(sf.sf_connect_in.value, true_name, MacOSEventType.EV_SFLT_CONNECT_IN, protocol=params["protocol"])
    if sf.sf_connect_out != 0:
        flag |= 1 << 9
        ql.os.ev_manager.register(sf.sf_connect_out.value, true_name, MacOSEventType.EV_SFLT_CONNECT_OUT, protocol=params["protocol"])
    if sf.sf_bind != 0:
        flag |= 1 << 10
        ql.os.ev_manager.register(sf.sf_bind.value, true_name, MacOSEventType.EV_SFLT_BIND, protocol=params["protocol"])
    if sf.sf_setoption != 0:
        flag |= 1 << 11
        ql.os.ev_manager.register(sf.sf_setoption.value, true_name, MacOSEventType.EV_SFLT_SETOPTION, protocol=params["protocol"])
    if sf.sf_getoption != 0:
        flag |= 1 << 12
        ql.os.ev_manager.register(sf.sf_getoption.value, true_name, MacOSEventType.EV_SFLT_GETOPTION, protocol=params["protocol"])
    if sf.sf_listen != 0:
        flag |= 1 << 13
        ql.os.ev_manager.register(sf.sf_listen.value, true_name, MacOSEventType.EV_SFLT_LISTEN, protocol=params["protocol"])
    if sf.sf_ioctl != 0:
        flag |= 1 << 14
        ql.os.ev_manager.register(sf.sf_ioctl.value, true_name, MacOSEventType.EV_SFLT_IOCTL, protocol=params["protocol"])

    ql.log.debug("New sflt callbacks has been registered: %s (%d/%d/%d/%d/%d/%d/%d/%d/%d/%d/%d/%d/%d/%d/%d)" % (
        true_name,
        (flag & (1 << 0)) != 0,
        (flag & (1 << 1)) != 0,
        (flag & (1 << 2)) != 0,
        (flag & (1 << 3)) != 0,
        (flag & (1 << 4)) != 0,
        (flag & (1 << 5)) != 0,
        (flag & (1 << 6)) != 0,
        (flag & (1 << 7)) != 0,
        (flag & (1 << 8)) != 0,
        (flag & (1 << 9)) != 0,
        (flag & (1 << 10)) != 0,
        (flag & (1 << 11)) != 0,
        (flag & (1 << 12)) != 0,
        (flag & (1 << 13)) != 0,
        (flag & (1 << 14)) != 0,))

    return 0

@macos_kernel_api(params={
    "handle": UINT,
})
def hook__sflt_unregister(ql, address, params):
    handle = str(params["handle"]).encode()

    evs = ql.os.ev_manager.get_events_by_name(b"", keyword=handle)
    for found in evs:
        ql.os.ev_manager.emit(found.name, MacOSEventType.EV_SFLT_UNREGISTERED, [params["handle"]])
        break

    ql.os.ev_manager.deregister(b"", keyword=handle)
    ql.log.debug("A sflt event has been deregistered: %s" % (handle))
    return 0

@macos_kernel_api(params={
    "mpc": POINTER,
    "handlep": POINTER,
    "xd": POINTER,
})
def hook__mac_policy_register(ql, address, params):
    mpc = mac_policy_conf_t(ql, params["mpc"])
    mpc = mpc.loadFromMem()

    mac_ops_addr = mpc.mpc_ops.value

    true_name = ql.mem.string(mpc.mpc_name.value).encode()
    for i in range(0, NUM_EVENT_MAC_POLICY):
        func, = struct.unpack("<Q", ql.mem.read(mac_ops_addr + i * 8, 8))
        if func != 0:
            ql.os.ev_manager.register(func, true_name, MACPolicy_EventType[MACPolicy_EventType(i + base_event_MAC).name])
            ql.log.debug("New mac policy callback has been registered: Name: %s --- Type: %s --- Addr: 0x%x " % 
                (true_name, MACPolicy_EventType(i + base_event_MAC).name, func))
            
    # Update mac_policy_list
    idx = ql.mac_policy_list.numloaded
    ql.mac_policy_list.numloaded += 1
    ql.mac_policy_list.updateToMem()

    if ql.mac_policy_list.entries.value != 0:
        entry = mac_policy_conf_t(ql, ql.mac_policy_list.entries.value)
        entry = entry.loadFromMem()

        while entry.mpc_list.value != 0:
            tmp = mac_policy_conf_t(ql, entry.mpc_list.value)
            tmp = tmp.loadFromMem()
            entry = tmp

        entry.mpc_list = POINTER64(params["mpc"])
        entry.updateToMem()
    else:
        ql.mac_policy_list.entries = POINTER64(params["mpc"])
        ql.mac_policy_list.updateToMem()

    ql.mem.write(params["handlep"], struct.pack("<I", idx))
    ql.policy_manager.bank[idx] = mpc.base

    # FIXME: Should place these before or after register?
    ql.os.ev_manager.emit_by_type(MACPolicy_EventType.EV_MAC_mpo_policy_init, [params["mpc"]])
    ql.os.ev_manager.emit_by_type(MACPolicy_EventType.EV_MAC_mpo_policy_initbsd, [params["mpc"]])
    
    return 0

@macos_kernel_api(params={
    "handle": POINTER
})
def hook__mac_policy_unregister(ql, address, params):
    handle = ql.policy_manager.bank[params["handle"]]
    mpc = mac_policy_conf_t(ql, handle)
    mpc = mpc.loadFromMem()

    ql.os.ev_manager.emit_by_type(MACPolicy_EventType.EV_MAC_mpo_policy_destroy, [mpc.base])

    true_name = ql.mem.string(mpc.mpc_name.value).encode()
    ql.os.ev_manager.deregister(true_name)
    del ql.policy_manager.bank[params["handle"]]

    # Update mac_policy_list
    entry = mac_policy_conf_t(ql, ql.mac_policy_list.entries.value)
    entry = entry.loadFromMem()

    if entry.base != handle:
        while entry.mpc_list.value != handle:
            tmp = mac_policy_conf_t(ql, entry.mpc_list.value)
            tmp = tmp.loadFromMem()
            entry = tmp

        entry.mpc_list = POINTER64(mpc.mpc_list.value)
    else:
        ql.mac_policy_list.entries = POINTER64(0)

    ql.log.debug("A MAC event has been deregistered: 0x%x" % handle)
    return 0;

@macos_kernel_api(params={
    "_identifier": POINTER,
    "_callback": POINTER,
    "_idata": POINTER,
})
def hook__kauth_listen_scope(ql, address, params):
    ev_name = ql.mem.string(params["_identifier"]).replace("com.", "").replace("apple.", "").upper().replace(".", "_")
    ql.os.ev_manager.register(params["_callback"], ev_name.encode(), MacOSEventType["EV_" + ev_name])
    ql.log.debug("New kauth callback has been registered: %s" % ev_name)
    return params["_identifier"]

@macos_kernel_api(params={
    "filter": POINTER,
    "filter_ref": POINTER,
})
def hook__ipf_addv4(ql, address, params):
    ip_filter_ipv4 = ipf_filter_t(ql, params["filter"]).loadFromMem()
    flag = 0
    name = ql.mem.string(ip_filter_ipv4.name.value).encode() + b"_ipf"
    cookie = ip_filter_ipv4.cookie.value

    ql.os.ev_manager.ipf_cookie[name] = cookie

    if ip_filter_ipv4.ipf_input.value > 0:
        flag |= (1 << 0)
        ql.os.ev_manager.register(ip_filter_ipv4.ipf_input.value, name, MacOSEventType.EV_IPF_INPUT)

    if ip_filter_ipv4.ipf_output.value > 0:
        flag |= (1 << 1)
        ql.os.ev_manager.register(ip_filter_ipv4.ipf_output.value, name, MacOSEventType.EV_IPF_OUTPUT)

    if ip_filter_ipv4.ipf_detach.value > 0:
        flag |= (1 << 2)
        ql.os.ev_manager.register(ip_filter_ipv4.ipf_detach.value, name, MacOSEventType.EV_IPF_DETACH)

    ql.log.debug("New ipf callbacks has been registered: %s (%d/%d/%d)" % (
        name,
        (flag & (1 << 0)) != 0,
        (flag & (1 << 1)) != 0,
        (flag & (1 << 2)) != 0,
        ))

    ql.mem.write(params["filter_ref"], struct.pack("<Q", ip_filter_ipv4.base))
    return 0;


##################################################################

########################## Syscall ###############################

@macos_kernel_api(params={
    "p": POINTER,
    "getattrlistbulk_args": POINTER,
    "retval": POINTER,
})
def hook__getattrlistbulk(ql, address, params):
    getattrlistbulk_args = getattrlistbulk_args_t(ql, params["getattrlistbulk_args"]).loadFromMem()
    dirfd = ql.os.ev_manager.map_fd[getattrlistbulk_args.dirfd]

    vfs_attr_pack = ql.loader.kernel_extrn_symbols_detail[b"_vfs_attr_pack"]["n_value"]

    uiovp_addr = ql.os.heap.alloc(ctypes.sizeof(user_iovec_t))
    uiovp = user_iovec_t(ql, uiovp_addr)
    uiovp.iov_base = getattrlistbulk_args.attributeBuffer
    uiovp.iov_len = getattrlistbulk_args.bufferSize
    uiovp.updateToMem()

    uio_addr = ql.os.heap.alloc(ctypes.sizeof(uio_t))
    uio = uio_t(ql, uio_addr)
    uio.uio_iovs = iovecs_t(kiovp=POINTER64(uiovp_addr), uiovp=POINTER64(uiovp_addr))
    uio.uio_iovcnt = 1
    uio.uio_offset = 0
    uio.uio_segflg = 8 # UIO_USERSPACE64
    uio.uio_rw = 0 # UIO_READ
    uio.uio_resid_64 = getattrlistbulk_args.bufferSize
    uio.uio_size = 72
    uio.uio_max_iovs = 1
    uio.uio_flags = 1
    uio.updateToMem()

    result = 0
    for path in dirfd.iterdir():
        result += 1
    ql.mem.write(params["retval"], struct.pack("<Q", result))

    for path in dirfd.iterdir():
        info = path.stat()

        vap_addr = ql.os.heap.alloc(ctypes.sizeof(vnode_attr_t))
        vap = vnode_attr_t(ql, vap_addr)
        vap.va_supported = 51573293058
        vap.va_active = 51573293058
        vap.va_nlink = info.st_nlink

        vap.va_total_size = info.st_size
        vap.va_data_size = info.st_size
        vap.va_uid = info.st_uid
        vap.va_gid = info.st_gid
        vap.va_mode = info.st_mode
        vap.va_fileid = info.st_ino
        vap.va_devid = info.st_dev
        vap.va_create_time = timespec_t(tv_sec = int(info.st_ctime), tv_nsec = info.st_ctime_ns % 1000000)
        vap.va_access_time = timespec_t(tv_sec = int(info.st_atime), tv_nsec = info.st_atime_ns % 1000000)
        vap.va_modify_time = timespec_t(tv_sec = int(info.st_mtime), tv_nsec = info.st_mtime_ns % 1000000)
        truename = path.name + "\x00"
        vap.va_name = POINTER64(ql.os.heap.alloc(1024))
        ql.mem.write(vap.va_name.value, truename.encode())
        vap.updateToMem()

        code = gen_stub_code(ql, [0, uio.base, getattrlistbulk_args.alist, getattrlistbulk_args.options, vap.base, 0, params["p"]], vfs_attr_pack)
        print("Trampoline created at 0x%x for %s (0x%x) and 0x%x" % (code, truename, vap.va_name.value, vap.base))
        ql.stack_push(code)

    return



##################################################################



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from qiling.os.windows.const import *
from qiling.os.windows.fncc import *


dllname = 'kernel32_dll'

# LONG InterlockedExchange(
#  LONG volatile *Target,
#  LONG          Value
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"Target": POINTER, "Value": UINT})
def hook_InterlockedExchange(ql, address, params):
    old = int.from_bytes(ql.mem.read(params['Target'], ql.pointersize), byteorder='little')
    ql.mem.write(params['Target'], params['Value'].to_bytes(length=ql.pointersize, byteorder='little'))
    return old


# LONG InterlockedIncrement(
#  LONG volatile *Target,
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"Target": POINTER})
def hook_InterlockedIncrement(ql, address, params):
    val = int.from_bytes(ql.mem.read(params['Target'], 4), byteorder='little')
    val += 1 & (2 ** 32)  # increment and overflow back to 0 if applicable
    ql.mem.write(params['Target'], val.to_bytes(length=4, byteorder='little'))
    return val


# LONG InterlockedDecrement(
#  LONG volatile *Target,
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"Target": POINTER})
def hook_InterlockedDecrement(ql, address, params):
    val = int.from_bytes(ql.mem.read(params['Target'], 4), byteorder='little')
    val -= 1
    if val == -1:
        val = 0xFFFFFFFF
    ql.mem.write(params['Target'], val.to_bytes(length=4, byteorder='little'))
    return val


# NTSYSAPI ULONGLONG VerSetConditionMask(
#   ULONGLONG ConditionMask, => 64bit param
#   DWORD     TypeMask,
#   BYTE      Condition
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"ConditionMask": ULONGLONG, "TypeMask": DWORD, "Condition": BYTE})
def hook_VerSetConditionMask(ql, address, params):
    # ConditionMask = params["ConditionMask"]
    TypeMask = params["TypeMask"]
    Condition = params["Condition"]
    mask = params["ConditionMask"]
    ConditionMask = ql.os.hooks_variables.get("ConditionMask", {})
    if TypeMask == 0:
        ret = ConditionMask
    else:
        Condition &= VER_CONDITION_MASK
        if Condition == 0:
            ret = ConditionMask
        else:
            ullCondMask = Condition
            if TypeMask & VER_PRODUCT_TYPE:
                # ConditionMask |= ullCondMask << (7 * VER_NUM_BITS_PER_CONDITION_MASK)
                ConditionMask[VER_PRODUCT_TYPE] = Condition
            elif TypeMask & VER_SUITENAME:
                # ConditionMask |= ullCondMask << (6 * VER_NUM_BITS_PER_CONDITION_MASK)
                ConditionMask[VER_SUITENAME] = Condition
            elif TypeMask & VER_SERVICEPACKMAJOR:
                # ConditionMask |= ullCondMask << (5 * VER_NUM_BITS_PER_CONDITION_MASK)
                ConditionMask[VER_SERVICEPACKMAJOR] = Condition
            elif TypeMask & VER_SERVICEPACKMINOR:
                # ConditionMask |= ullCondMask << (4 * VER_NUM_BITS_PER_CONDITION_MASK)
                ConditionMask[VER_SERVICEPACKMINOR] = Condition
            elif TypeMask & VER_PLATFORMID:
                # ConditionMask |= ullCondMask << (3 * VER_NUM_BITS_PER_CONDITION_MASK)
                ConditionMask[VER_PLATFORMID] = Condition
            elif TypeMask & VER_BUILDNUMBER:
                # ConditionMask |= ullCondMask << (2 * VER_NUM_BITS_PER_CONDITION_MASK)
                ConditionMask[VER_BUILDNUMBER] = Condition
            elif TypeMask & VER_MAJORVERSION:
                # ConditionMask |= ullCondMask << (1 * VER_NUM_BITS_PER_CONDITION_MASK)
                ConditionMask[VER_MAJORVERSION] = Condition
            elif TypeMask & VER_MINORVERSION:
                # ConditionMask |= ullCondMask << (0 * VER_NUM_BITS_PER_CONDITION_MASK)
                ConditionMask[VER_MINORVERSION] = Condition
            ret = 1
    # ConditionMask should be updated locally
    # https://docs.microsoft.com/it-it/windows/win32/sysinfo/verifying-the-system-version
    # But since we don't have the pointer to the variable, an hack is to use the environment.
    # Feel free to push a better solution
    # Since I can't work with bits, and since we had to work with the environment anyway, let's use a dict
    ql.os.hooks_variables["ConditionMask"] = ConditionMask
    return ret



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os


from qiling.const import *

def ql_syscall_getrandom(ql, buf, buflen, flags,*args, **kw):
    data = None
    regreturn = None
    try:
        data = os.urandom(buflen)
        ql.uc.mem_write(buf, data)
        regreturn = len(data)
    except:
        regreturn = -1

    if data:
        ql.log.debug("getrandom() CONTENT:")
        ql.log.debug(str(data))
    return regreturn



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct
import time
from qiling.os.windows.const import *
from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *


dllname = 'kernel32_dll'

# DWORD FlsFree(
#  DWORD dwFlsIndex
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'DWORD': 'UINT'})
def hook_FlsFree(ql, address, params):
    return ql.os.fiber_manager.free(params['dwFlsIndex'])


# LPVOID FlsGetValue(
#  DWORD dwFlsIndex
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'DWORD': 'UINT'})
def hook_FlsGetValue(ql, address, params):
    return ql.os.fiber_manager.get(params['dwFlsIndex'])


# LPVOID FlsSetValue(
#  DWORD dwFlsIndex
#  PVOID lpFlsData
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'DWORD': 'UINT'})
def hook_FlsSetValue(ql, address, params):
    return ql.os.fiber_manager.set(params['dwFlsIndex'], params['lpFlsData'])


# DWORD FlsAlloc(
#  PFLS_CALLBACK_FUNCTION lpCallback
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_FlsAlloc(ql, address, params):
    # global cb = params['lpCallback']
    cb = params['lpCallback']
    if cb:
        return ql.os.fiber_manager.alloc(cb)
    else:
        return ql.os.fiber_manager.alloc()
